"""This module holds shim backends used for smoke tests."""
import json
from mellea.backends import Backend, BaseModelSubclass
from mellea.stdlib.base import CBlock, Component, Context, ModelOutputThunk, GenerateLog
from mellea.backends.formatter import Formatter, TemplateFormatter
from ..llm.i_tg_llm import I_TG_LLM


class SimpleBackend(Backend):

    formatter: Formatter
    llm: I_TG_LLM

    def __init__(self, llm: I_TG_LLM):
        self.llm = llm
        self.formatter = TemplateFormatter(model_id=llm.model_name)
    
    async def generate_from_context(
        self,
        action: Component | CBlock,
        ctx: Context,
        *,
        format: type[BaseModelSubclass] | None = None,
        model_options: dict | None = None,
        tool_calls: bool = False,
    ) -> tuple[ModelOutputThunk, Context]:
        
        msg = self.formatter.to_chat_messages([action])[0]
        msg = {
            "role": msg.role,
            "content": [{"type": "text", "text": msg.content}],
        }

        resp = await self.llm.generate([msg])
        
        res = {"result": resp}
        mot = ModelOutputThunk(value=json.dumps(res))
        mot._generate_log = GenerateLog()
        return mot, ctx.add(action).add(mot)
    
    async def generate_from_raw(
        self,
        actions: list[Component | CBlock],
        ctx: Context,
        *,
        format: type[BaseModelSubclass] | None = None,
        model_options: dict | None = None,
        tool_calls: bool = False,
    ) -> list[ModelOutputThunk]:
        
        raise NotImplementedError()


from typing import List
from smolagents.local_python_executor import LocalPythonExecutor

def run_safe_python(code: str, libs: List[str] = []):
    exec = LocalPythonExecutor(
        additional_authorized_imports = libs,
        max_print_outputs_length= None,
        additional_functions = None,
    )
    out = exec(code)
    return out


from ..common.str import to_snake_case
from ..data_types import ToolGuardSpec, ToolGuardSpecItem


RUNTIME_PACKAGE_NAME = "rt_toolguard"
RUNTIME_INIT_PY = "__init__.py"
RUNTIME_TYPES_PY = "data_types.py"
RUNTIME_APP_TYPES_PY = "domain_types.py"

def guard_fn_name(tool_policy: ToolGuardSpec) -> str:
    return to_snake_case(f"guard_{tool_policy.tool_name}")


def guard_fn_module_name(tool_policy: ToolGuardSpec) -> str:
    return to_snake_case(f"guard_{tool_policy.tool_name}")


def guard_item_fn_name(tool_item: ToolGuardSpecItem) -> str:
    return to_snake_case(f"guard_{tool_item.name}")


def guard_item_fn_module_name(tool_item: ToolGuardSpecItem) -> str:
    return to_snake_case(f"guard_{tool_item.name}")


def test_fn_name(tool_item: ToolGuardSpecItem) -> str:
    return to_snake_case(f"test_guard_{tool_item.name}")


def test_fn_module_name(tool_item: ToolGuardSpecItem) -> str:
    return to_snake_case(f"test_guard_{tool_item.name}")



import inspect
import re
from typing import Callable


def extract_docstr_args(func: Callable) -> str:
    doc = inspect.getdoc(func)
    if not doc:
        return ""

    lines = doc.splitlines()

    def args_start_line():
        for i, line in enumerate(lines):
            if line.strip().lower() == "args:":  # Google style docstr
                return i + 1
            if (
                line.strip().lower().startswith(":param ")
            ):  # Sphinx-style docstring. https://sphinx-rtd-tutorial.readthedocs.io/en/latest/docstrings.html
                return i

    args_start = args_start_line()
    if args_start is None:
        return ""

    # List of known docstring section headers
    next_sections = {
        "returns:",
        "raises:",
        "examples:",
        "notes:",
        "attributes:",
        "yields:",
    }

    # Capture lines after "Args:" that are indented
    args_lines = []
    indent = " " * 4 * 2
    for line in lines[args_start:]:
        # Stop if we hit a new section (like "Returns:", "Raises:", etc.)
        stripped = line.strip().lower()
        if stripped in next_sections or stripped.startswith(":return:"):
            break

        args_lines.append(indent + sphinx_param_to_google(line.strip()))

    # Join all lines into a single string
    if not args_lines:
        return ""

    return "\n".join(args_lines)


def sphinx_param_to_google(line: str) -> str:
    """
    Convert a single Sphinx-style ':param' line to Google style.

    Args:
        line: A Sphinx param line, e.g.
              ':param user_id: The unique identifier of the user.'

    Returns:
        str: Google style equivalent, e.g.
             'user_id: The unique identifier of the user.'
    """
    m = re.match(r"\s*:param\s+(\w+)\s*:\s*(.*)", line)
    if not m:
        return line
    name, desc = m.groups()
    return f"{name}: {desc}"



import sys
import importlib
from pathlib import Path
import multiprocessing

def process_target(queue, func, args, kwargs, extra_paths):
    try:
        # Add dynamic python paths
        if extra_paths:
            for p in extra_paths:
                p = str(Path(p).resolve())
                if p not in sys.path:
                    sys.path.insert(0, p)

        # Force reload of modules from those paths (optional)
        if extra_paths:
            for modname, module in list(sys.modules.items()):
                file = getattr(module, "__file__", None)
                if not file:
                    continue

                for p in extra_paths:
                    p = Path(p).resolve()
                    try:
                        if Path(file).resolve().is_relative_to(p):
                            importlib.reload(module)
                    except Exception:
                        pass

        # Run the function
        result = func(*args, **kwargs)
        queue.put(("ok", result))
    except Exception as e:
        queue.put(("error", e))

def run_in_process(func, *args, extra_paths=None, **kwargs):
    queue = multiprocessing.Queue()
    p = multiprocessing.Process(
        target=process_target,
        args=(queue, func, args, kwargs, extra_paths)
    )
    p.start()
    p.join()

    status, value = queue.get()
    if status == "error":
        raise value
    return value



import re
from typing import Set
from mellea import MelleaSession
from ..data_types import Domain, ToolGuardSpecItem
from .prompts.pseudo_code import tool_policy_pseudo_code

MAX_TRIALS = 3


async def tool_dependencies(
    policy_item: ToolGuardSpecItem, tool_signature: str, domain: Domain, m: MelleaSession, trial=0
) -> Set[str]:
    model_options = {}#{ModelOption.TEMPERATURE: 0.8}
    pseudo_code = tool_policy_pseudo_code(
        m,
        policy_item=policy_item,
        fn_to_analyze=tool_signature,
        domain=domain,
        model_options=model_options,
    )  # type: ignore
    fn_names = _extract_api_calls(pseudo_code)
    if all([f"{fn_name}(" in domain.app_api.content for fn_name in fn_names]):
        return fn_names
    if trial <= MAX_TRIALS:
        # as tool_policy_pseudo_code has some temerature, we retry hoping next time the pseudo code will be correct
        return await tool_dependencies(policy_item, tool_signature, domain, m, trial + 1)
    raise Exception("Failed to analyze api dependencies")


def _extract_api_calls(code: str) -> Set[str]:
    pattern = re.compile(r"\bapi\.(\w+)\s*\(")
    return set(pattern.findall(code))


import json
import sys
import subprocess
from pydantic import BaseModel
from typing import List, Optional

from ...data_types import FileTwin

ERROR = "error"
WARNING = "warning"


class Position(BaseModel):
    line: int
    character: int


class Range(BaseModel):
    start: Position
    end: Position


class GeneralDiagnostic(BaseModel):
    file: str
    severity: str
    message: str
    range: Range
    rule: Optional[str] = None


class Summary(BaseModel):
    filesAnalyzed: int
    errorCount: int
    warningCount: int
    informationCount: int
    timeInSec: float


class DiagnosticsReport(BaseModel):
    version: str
    time: str
    generalDiagnostics: List[GeneralDiagnostic]
    summary: Summary

    def list_error_messages(self, file_content: str) -> List[str]:
        msgs = set()
        for d in self.generalDiagnostics:
            if d.severity == ERROR:
                msgs.add(
                    f"Syntax error: {d.message}.  code block: '{get_text_by_range(file_content, d.range)}, '"
                )
        return list(msgs)


def get_text_by_range(file_content: str, rng: Range) -> str:
    lines = file_content.splitlines()

    if rng.start.line == rng.end.line:
        # Single-line span
        return lines[rng.start.line][rng.start.character : rng.end.character]

    # Multi-line span
    selected_lines = []
    selected_lines.append(
        lines[rng.start.line][rng.start.character :]
    )  # First line, from start.character
    for line_num in range(rng.start.line + 1, rng.end.line):
        selected_lines.append(lines[line_num])  # Full middle lines
    selected_lines.append(
        lines[rng.end.line][: rng.end.character]
    )  # Last line, up to end.character

    return "\n".join(selected_lines)


def run(folder: str, py_file: str) -> DiagnosticsReport:
    py_path = sys.executable
    res = subprocess.run(
        [
            "pyright",
            # "--venv-path", venv_path,
            "--pythonpath",
            py_path,
            "--outputjson",
            py_file,
        ],
        cwd=folder,
        capture_output=True,
        text=True,
    )
    # if res.returncode !=0:
    #     raise Exception(res.stderr)

    data = json.loads(res.stdout)
    return DiagnosticsReport.model_validate(data)


def config(folder: str):
    cfg = {
        "typeCheckingMode": "basic",
        "reportOptionalIterable": WARNING,
        "reportArgumentType": WARNING,  # "Object of type \"None\" cannot be used as iterable value",
        "reportOptionalMemberAccess": WARNING,
        "reportOptionalSubscript": WARNING,
        "reportAttributeAccessIssue": ERROR,
    }
    FileTwin(file_name="pyrightconfig.json", content=json.dumps(cfg, indent=2)).save(
        folder
    )



from abc import ABC, abstractmethod
from typing import List, Dict


class I_TG_LLM(ABC):
    model_name: str

    @abstractmethod
    async def chat_json(self, messages: List[Dict]) -> Dict:
        pass

    @abstractmethod
    async def generate(self, messages: List[Dict]) -> str:
        pass





from typing import List
from langchain_core.tools import BaseTool

from .open_api import OpenAPI

def langchain_tools_to_openapi(
    tools: List[BaseTool],
    title: str = "LangChain Tools API",
    version: str = "1.0.0",
)->OpenAPI:
    paths = {}
    components = {"schemas": {}}

    for tool in tools:
        # Get JSON schema from the args model
        if tool.get_input_schema():
            components["schemas"][tool.name + "Args"] = tool.get_input_schema().model_json_schema()

            request_body = {
                "description": tool.description,
                "required": True,
                "content": {
                    "application/json": {
                        "schema": {"$ref": f"#/components/schemas/{tool.name}Args"}
                    }
                },
            }
        else:
            # Tools without args → empty schema
            request_body = None

        paths[f"/tools/{tool.name}"] = {
            "post": {
                "summary": tool.description,
                "operationId": tool.name,
                "requestBody": request_body,
                "responses": {
                    "200": {
                        "description": "Tool result",
                        "content": {"application/json": tool.get_output_jsonschema()},
                    }
                },
            }
        }

    return OpenAPI.model_validate({
        "openapi": "3.1.0",
        "info": {"title": title, "version": version},
        "paths": paths,
        "components": components,
    })



from typing import Dict, Any
import os

import json
from typing import List


def read_prompt_file(filename: str) -> str:
    with open(
        os.path.join(os.path.dirname(__file__), "prompts", filename + ".txt"), "r"
    ) as f:
        return f.read()


def generate_messages(system_prompt: str, user_content: str) -> List[Dict[str, str]]:
    return [
        {"role": "system", "content": system_prompt},
        {"role": "user", "content": user_content},
    ]


def save_output(outdir: str, filename: str, content: Any):
    with open(os.path.join(outdir, filename), "w") as outfile:
        json.dump(content, outfile, indent=4)


def normalize_text(text):
    """Normalize text by removing punctuation, converting to lowercase, and standardizing spaces."""
    # return re.sub(r'\s+', ' ', re.sub(r'[^a-zA-Z0-9\s]', '', text)).strip().lower()
    return text.lower()


def split_reference_if_both_parts_exist(reference, policy_text):
    words = reference.split()
    for split_point in range(1, len(words)):
        part1 = " ".join(words[:split_point])
        part2 = " ".join(words[split_point:])

        normalized_part1 = normalize_text(part1)
        normalized_part2 = normalize_text(part2)
        normalized_policy_text = normalize_text(policy_text)

        if (
            normalized_part1 in normalized_policy_text
            and normalized_part2 in normalized_policy_text
        ):
            start_idx1 = normalized_policy_text.find(normalized_part1)
            end_idx1 = start_idx1 + len(part1)
            start_idx2 = normalized_policy_text.find(normalized_part2)
            end_idx2 = start_idx2 + len(part2)
            return [policy_text[start_idx1:end_idx1], policy_text[start_idx2:end_idx2]]
    return None


def find_mismatched_references(policy_text, policy_json):
    corrections = json.loads(json.dumps(policy_json))
    unmatched_policies = []
    if isinstance(corrections["policy_items"], str):
        return corrections, unmatched_policies

    normalized_policy_text = normalize_text(policy_text)

    for policy in corrections["policy_items"]:
        corrected_references = []
        has_unmatched = False

        for reference in policy["references"]:
            normalized_ref = normalize_text(reference)
            # if normalized ref in policy doc- just copy the original
            if normalized_ref in normalized_policy_text:
                start_idx = normalized_policy_text.find(normalized_ref)
                end_idx = start_idx + len(reference)
                corrected_references.append(policy_text[start_idx:end_idx])
            else:
                # close_match = get_close_matches(normalized_ref, [normalized_policy_text], n=1, cutoff=0.9)
                # if close_match:
                # 	start_idx = normalized_policy_text.find(close_match[0])
                # 	end_idx = start_idx + len(close_match[0])
                # 	corrected_references.append(policy_text[start_idx:end_idx])
                # else:
                split_segments = split_reference_if_both_parts_exist(
                    reference, policy_text
                )
                if split_segments:
                    corrected_references.extend(split_segments)
                else:
                    corrected_references.append(
                        reference
                    )  # Keep original if no match found
                    has_unmatched = True

        policy["references"] = corrected_references
        if has_unmatched:
            unmatched_policies.append(policy["name"])

    return corrections, unmatched_policies


#important to load the env variables BEFORE policy_adherence library (so programmatic_ai configuration will take place)
import dotenv
dotenv.load_dotenv()

from .cli import main

if __name__ == "__main__":
    main()


from dataclasses import is_dataclass
from enum import Enum
import inspect
import os
import textwrap
from types import FunctionType, UnionType
import types
from typing import (
    Callable,
    DefaultDict,
    Dict,
    List,
    Literal,
    Optional,
    Set,
    Tuple,
    get_type_hints,
    get_origin,
    get_args,
)
from typing import Annotated, Union
from collections import defaultdict, deque
import typing
from ..common.py import module_to_path
from ..data_types import FileTwin

Dependencies = DefaultDict[type, Set[type]]

class APIExtractor:
    def __init__(self, py_path:str, include_module_roots:List[str] = []):
        self.py_path = py_path
        self.include_module_roots = include_module_roots

    def extract_from_functions(
        self,
        funcs: List[Callable],
        interface_name: str,
        interface_module_name: str,
        types_module_name: str,
        impl_module_name: str,
        impl_class_name: str,
    ) -> Tuple[FileTwin, FileTwin, FileTwin]:
        assert all([_is_global_or_class_function(func) for func in funcs])

        os.makedirs(self.py_path, exist_ok=True)

        # used types
        types = FileTwin(
            file_name=module_to_path(types_module_name),
            content=self._generate_types_file(
                *self._collect_all_types_from_functions(funcs)
            ),
        ).save(self.py_path)

        # API interface
        interface = FileTwin(
            file_name=module_to_path(interface_module_name),
            content=self._generate_interface_from_functions(
                funcs, interface_name, types_module_name
            ),
        ).save(self.py_path)

        # API impl interface
        impl = FileTwin(
            file_name=module_to_path(impl_module_name),
            content=self._generate_impl_from_functions(
                funcs,
                impl_class_name,
                interface_module_name,
                interface_name,
                types_module_name,
            ),
        ).save(self.py_path)

        return interface, types, impl

    def extract_from_class(
        self,
        typ: type,
        *,
        interface_name: Optional[str] = None,
        interface_module_name: Optional[str] = None,
        types_module_name: Optional[str] = None,
    ) -> Tuple[FileTwin, FileTwin]:
        """Extract interface and types from a class and save to files."""
        class_name = _get_type_name(typ)
        interface_name = interface_name or "I_" + class_name
        interface_module_name = interface_module_name or f"I_{class_name}".lower()
        types_module_name = types_module_name or f"{class_name}_types".lower()

        os.makedirs(self.py_path, exist_ok=True)

        # Types
        collected, dependencies = self._collect_all_types_from_class(typ)
        types_content = self._generate_types_file(collected, dependencies)
        types_file = module_to_path(types_module_name)
        types = FileTwin(file_name=types_file, content=types_content).save(self.py_path)

        # API interface
        if_content = self._generate_interface_from_class(
            typ, interface_name, types_module_name
        )
        if_file = module_to_path(interface_module_name)
        interface = FileTwin(file_name=if_file, content=if_content).save(self.py_path)

        return interface, types

    def _generate_interface_from_class(
        self, typ: type, interface_name: str, types_module: str
    ) -> str:
        # Start building the interface
        lines = [
            "# Auto-generated class interface",
            "from typing import *  # type: ignore",
            "from abc import ABC, abstractmethod",
            f"from {types_module} import *",
            "",
        ]

        lines.append(f"class {interface_name}(ABC):")  # Abstract class

        # Add class docstring if available
        if typ.__doc__:
            docstring = typ.__doc__.strip()
            if docstring:
                lines.append('    """')
                # Handle multi-line docstrings
                for line in docstring.split("\n"):
                    lines.append(f"    {line.strip()}")
                lines.append('    """')

        # Get all methods
        methods = []
        for name, method in inspect.getmembers(typ, predicate=inspect.isfunction):
            if not name.startswith("_"):
                methods.append((name, method))

        if not methods:
            lines.append("    pass")
        else:
            for method_name, method in methods:
                # Add method docstring and signature
                lines.append("    @abstractmethod")
                method_lines = self._get_function_with_docstring(method, method_name)
                lines.extend([line if line else "" for line in method_lines])
                lines.append("        ...")
                lines.append("")

        return textwrap.dedent("\n".join(lines))

    def _generate_interface_from_functions(
        self, funcs: List[Callable], interface_name: str, types_module: str
    ) -> str:
        lines = [
            "# Auto-generated class interface",
            "from typing import * # type: ignore",
            "from abc import ABC, abstractmethod",
            f"from {types_module} import *",
            "",
        ]

        lines.append(f"class {interface_name}(ABC):")  # Abstract class
        lines.append("")

        indent = " " * 4
        if not funcs:
            lines.append(f"{indent}pass")
        else:
            for func in funcs:
                # Add method docstring and signature
                lines.append(f"{indent}@abstractmethod")
                method_lines = self._get_function_with_docstring(
                    func, _get_type_name(func)
                )
                lines.extend([line if line else "" for line in method_lines])
                lines.append(f"{indent}{indent}...")
                lines.append("")

        if any(["Decimal" in line for line in lines]):
            lines.insert(2, "from decimal import Decimal")

        return "\n".join(lines)

    def _generate_impl_from_functions(
        self,
        funcs: List[Callable],
        class_name: str,
        interface_module_name: str,
        interface_name: str,
        types_module: str,
    ) -> str:
        lines = [
            "# Auto-generated class",
            "from typing import *",
            "from abc import ABC, abstractmethod",
            f"from {interface_module_name} import {interface_name}",
            f"from {types_module} import *",
            "",
            """class IToolInvoker(ABC):
    T = TypeVar("T")
    @abstractmethod
    def invoke(self, toolname: str, arguments: Dict[str, Any], model: Type[T])->T:
        ...""",
            "",
        ]

        lines.append(f"class {class_name}({interface_name}):")  # class
        lines.append("")
        lines.append("""    def __init__(self, delegate: IToolInvoker):
        self._delegate = delegate
    """)

        if not funcs:
            lines.append("    pass")
        else:
            for func in funcs:
                # Add method docstring and signature
                method_lines = self._get_function_with_docstring(
                    func, _get_type_name(func)
                )
                lines.extend([line if line else "" for line in method_lines])
                lines.extend(self._generate_delegate_code(func))
                lines.append("")

        if any(["Decimal" in line for line in lines]):
            lines.insert(2, "from decimal import Decimal")

        return "\n".join(lines)

    def _generate_delegate_code(self, func: Callable) -> List[str]:
        func_name = _get_type_name(func)
        indent = " " * 4 * 2
        sig = inspect.signature(func)
        ret = sig.return_annotation
        if ret is inspect._empty:
            ret_name = "None"
        elif hasattr(ret, "__name__"):
            ret_name = ret.__name__
        else:
            ret_name = str(ret)
        return [
            indent + "args = locals().copy()",
            indent + "args.pop('self', None)",
            indent + f"return self._delegate.invoke('{func_name}', args, {ret_name})",
        ]

    def _get_function_with_docstring(
        self, func: FunctionType, func_name: str
    ) -> List[str]:
        """Extract method signature with type hints and docstring."""
        lines = []

        # Get method signature
        method_signature = self._get_method_signature(func, func_name)
        lines.append(f"    {method_signature}:")

        # Add method docstring if available
        if func.__doc__:
            docstring = func.__doc__
            indent = " " * 8
            if docstring:
                lines.append(indent + '"""')
                lines.extend(docstring.strip("\n").split("\n"))
                lines.append(indent + '"""')

        return lines

    def should_include_type(self, typ: type) -> bool:
        if hasattr(typ, "__module__"):
            module_root = typ.__module__.split(".")[0]
            if module_root in self.include_module_roots:
                return True
        return any([self.should_include_type(arg) for arg in get_args(typ)])

    def _generate_class_definition(self, typ: type) -> List[str]:
        """Generate a class definition with its fields."""
        lines = []
        class_name = _get_type_name(typ)

        if is_dataclass(typ):
            lines.append("@dataclass")

        # Determine base classes
        bases = [_get_type_name(b) for b in _get_type_bases(typ)]
        inheritance = f"({', '.join(bases)})" if bases else ""
        lines.append(f"class {class_name}{inheritance}:")

        # #is Pydantic?
        # is_pydantic = False
        # for base in cls.__bases__:
        #     if hasattr(base, '__module__') and 'pydantic' in str(base.__module__):
        #         is_pydantic = True

        indent = " " * 4
        # Add class docstring if available
        if typ.__doc__:
            docstring = typ.__doc__
            if docstring:
                lines.append(f'{indent}"""')
                lines.extend(
                    [f"{indent}{line}" for line in docstring.strip("\n").split("\n")]
                )
                lines.append(f'{indent}"""')

        # Fields
        annotations = getattr(typ, "__annotations__", {})
        if annotations:
            field_descriptions = self._extract_field_descriptions(typ)
            for field_name, field_type in annotations.items():
                if field_name.startswith("_"):
                    continue

                # Handle optional field detection by default=None
                is_optional = False
                default_val = getattr(typ, field_name, ...)
                if default_val is None:
                    is_optional = True
                elif hasattr(typ, "__fields__"):
                    # Pydantic field with default=None
                    field_info = typ.__fields__.get(field_name)
                    if field_info and field_info.is_required() is False:
                        is_optional = True

                type_str = self._format_type(field_type)

                # Avoid wrapping Optional twice
                if is_optional:
                    origin = get_origin(field_type)
                    args = get_args(field_type)
                    already_optional = (
                        origin is typing.Union
                        and type(None) in args
                        or type_str.startswith("Optional[")
                    )
                    if not already_optional:
                        type_str = f"Optional[{type_str}]"

                # Check if we have a description for this field
                description = field_descriptions.get(field_name)

                # if description and is_pydantic:
                #     # Use Pydantic Field with description
                #     lines.append(f'    {field_name}: {type_str} = Field(description="{description}")')
                if description:
                    # Add description as comment for non-Pydantic classes
                    lines.append(f"{indent}{field_name}: {type_str}  # {description}")
                else:
                    # No description available
                    lines.append(f"{indent}{field_name}: {type_str}")

        # Enum
        elif issubclass(typ, Enum):
            if issubclass(typ, str):
                lines.extend(
                    [f'{indent}{entry.name} = "{entry.value}"' for entry in typ]
                )
            else:
                lines.extend([f"{indent}{entry.name} = {entry.value}" for entry in typ])

        else:
            lines.append(f"{indent}pass")

        return lines

    def _extract_field_descriptions(self, typ: type) -> Dict[str, str]:
        """Extract field descriptions from various sources."""
        descriptions = {}

        # Method 1: Check for Pydantic Field definitions
        if hasattr(typ, "__fields__"):  # Pydantic v1
            for field_name, field_info in typ.__fields__.items():
                if hasattr(field_info, "field_info") and hasattr(
                    field_info.field_info, "description"
                ):
                    descriptions[field_name] = field_info.field_info.description
                elif hasattr(field_info, "description") and field_info.description:
                    descriptions[field_name] = field_info.description

        # Method 2: Check for Pydantic v2 model fields
        if hasattr(typ, "model_fields"):  # Pydantic v2
            for field_name, field_info in typ.model_fields.items():
                if hasattr(field_info, "description") and field_info.description:
                    descriptions[field_name] = field_info.description

        # Method 3: Check class attributes for Field() definitions
        for attr_name in dir(typ):
            if not attr_name.startswith("_"):
                try:
                    attr_value = getattr(typ, attr_name)
                    # Check if it's a Pydantic Field
                    if hasattr(attr_value, "description") and attr_value.description:
                        descriptions[attr_name] = attr_value.description
                    elif hasattr(attr_value, "field_info") and hasattr(
                        attr_value.field_info, "description"
                    ):
                        descriptions[attr_name] = attr_value.field_info.description
                except Exception:
                    pass

        # Method 4: Parse class source for inline comments or docstrings
        try:
            source_lines = inspect.getsourcelines(typ)[0]
            current_field = None

            for line in source_lines:
                line = line.strip()

                # Look for field definitions with type hints
                if (
                    ":" in line
                    and not line.startswith("def ")
                    and not line.startswith("class ")
                ):
                    # Extract field name
                    field_part = line.split(":")[0].strip()
                    if " " not in field_part and field_part.isidentifier():
                        current_field = field_part

                # Look for comments on the same line or next line
                if "#" in line and current_field:
                    comment = line.split("#", 1)[1].strip()
                    if comment and current_field not in descriptions:
                        descriptions[current_field] = comment
                    current_field = None

        except Exception:
            pass

        # Method 5: Check for dataclass field descriptions
        if hasattr(typ, "__dataclass_fields__"):
            for field_name, field in typ.__dataclass_fields__.items():
                if hasattr(field, "metadata") and "description" in field.metadata:
                    descriptions[field_name] = field.metadata["description"]

        return descriptions

    def _get_method_signature(self, method: FunctionType, method_name: str):
        """Extract method signature with type hints."""
        try:
            sig = inspect.signature(method)
            # Get param hints
            try:
                param_hints = get_type_hints(method)
            except Exception:
                param_hints = {}

            params = []
            if not sig.parameters.get("self"):
                params.append("self")

            for param_name, param in sig.parameters.items():
                param_str = param_name

                # Add type annotation if available
                if param_name in param_hints:
                    type_str = self._format_type(param_hints[param_name])
                    param_str += f": {type_str}"
                elif param.annotation != param.empty:
                    param_str += f": {param.annotation}"

                # Add default value if present
                if param.default != param.empty:
                    if isinstance(param.default, str):
                        param_str += f' = "{param.default}"'
                    else:
                        param_str += f" = {repr(param.default)}"

                params.append(param_str)

            # Handle return type
            return_annotation = ""
            if "return" in param_hints:
                if param_hints["return"] is not type(None):
                    return_type = self._format_type(param_hints["return"])
                    return_annotation = f" -> {return_type}"
            elif sig.return_annotation != sig.empty:
                return_annotation = f" -> {sig.return_annotation}"

            params_str = ", ".join(params)
            return f"def {method_name}({params_str}){return_annotation}"

        except Exception:
            # Fallback for problematic signatures
            return f"def {method_name}(self, *args, **kwargs)"

    def _collect_all_types_from_functions(
        self, funcs: List[Callable]
    ) -> Tuple[Set[type], Dependencies]:
        processed_types = set()
        collected = set()
        dependencies = defaultdict(set)

        for func in funcs:
            for param, hint in get_type_hints(func).items():  # noqa: B007
                self._collect_types_recursive(
                    hint, processed_types, collected, dependencies
                )

        return collected, dependencies

    def _collect_all_types_from_class(
        self, typ: type
    ) -> Tuple[Set[type], Dependencies]:
        """Collect all types used in the class recursively."""
        visited = set()
        collected = set()
        dependencies = defaultdict(set)

        # Field types
        try:
            class_hints = get_type_hints(typ)
            for field, hint in class_hints.items():  # noqa: B007
                self._collect_types_recursive(hint, visited, collected, dependencies)
        except Exception:
            pass

        # Methods and param types
        for name, method in inspect.getmembers(typ, predicate=inspect.isfunction):  # noqa: B007
            try:
                method_hints = get_type_hints(method)
                for hint in method_hints.values():
                    self._collect_types_recursive(
                        hint, visited, collected, dependencies
                    )
            except Exception:
                pass

        # Also collect base class types
        for base in _get_type_bases(typ):
            self._collect_types_recursive(base, visited, collected, dependencies)

        return collected, dependencies

    def _collect_types_recursive(
        self, typ: type, visited: Set[type], acc: Set[type], dependencies: Dependencies
    ):
        """Recursively collect all types from a type hint."""
        visited.add(typ)

        if not self.should_include_type(typ):
            return

        acc.add(typ)
        origin = get_origin(typ)
        args = get_args(typ)

        # Type with generic arguments. eg: List[Person]
        if origin and args:
            for f_arg in args:
                self._collect_types_recursive(f_arg, visited, acc, dependencies)
                self._add_dependency(typ, f_arg, dependencies)
            return

        # If it's a custom class, try to get its type hints
        try:
            field_hints = typ.__annotations__  # direct fields
            for field_name, field_hint in field_hints.items():  # noqa: B007
                f_origin = get_origin(field_hint)
                if f_origin:
                    for f_arg in get_args(field_hint):
                        self._collect_types_recursive(f_arg, visited, acc, dependencies)
                        self._add_dependency(typ, f_arg, dependencies)
                else:
                    self._collect_types_recursive(
                        field_hint, visited, acc, dependencies
                    )
                    self._add_dependency(typ, field_hint, dependencies)

            for base in _get_type_bases(typ):  # Base classes
                self._collect_types_recursive(base, visited, acc, dependencies)
                self._add_dependency(typ, base, dependencies)
        except Exception:
            pass

    def _add_dependency(
        self, dependent_type: type, dependency_type: type, dependencies: Dependencies
    ):
        """Add a dependency relationship between types."""
        dep_name = _get_type_name(dependent_type)
        dep_on_name = _get_type_name(dependency_type)
        if dep_name != dep_on_name:
            dependencies[dependent_type].add(dependency_type)

        for arg in get_args(dependency_type):
            dependencies[dependent_type].add(arg)

    def _topological_sort_types(self, types: List[type], dependencies: Dependencies):
        """Sort types by their dependencies using topological sort."""
        # Create a mapping of type names to types for easier lookup
        type_map = {_get_type_name(t): t for t in types}

        # Build adjacency list and in-degree count
        adj_list = defaultdict(list)
        in_degree = defaultdict(int)

        # Initialize in-degree for all types
        for t in types:
            type_name = _get_type_name(t)
            if type_name not in in_degree:
                in_degree[type_name] = 0

        # Build the dependency graph
        for dependent_type in types:
            dependent_name = _get_type_name(dependent_type)
            for dependency_type in dependencies.get(dependent_type, set()):
                dependency_name = _get_type_name(dependency_type)
                if (
                    dependency_name in type_map
                ):  # Only consider types we're actually processing
                    adj_list[dependency_name].append(dependent_name)
                    in_degree[dependent_name] += 1

        # Kahn's algorithm for topological sorting
        queue = deque([name for name in in_degree if in_degree[name] == 0])
        result = []

        while queue:
            current = queue.popleft()
            result.append(type_map[current])

            for neighbor in adj_list[current]:
                in_degree[neighbor] -= 1
                if in_degree[neighbor] == 0:
                    queue.append(neighbor)

        # If we couldn't sort all types, there might be circular dependencies
        # Add remaining types at the end
        sorted_names = {_get_type_name(t) for t in result}
        remaining = [t for t in types if _get_type_name(t) not in sorted_names]
        result.extend(remaining)

        return result

    def _generate_types_file(
        self, collected_types: Set[type], dependencies: Dependencies
    ) -> str:
        """Generate the types file content."""
        lines = []
        lines.append("# Auto-generated type definitions")
        lines.append("from datetime import date, datetime")
        lines.append("from enum import Enum")
        lines.append("from typing import *")
        lines.append("from pydantic import BaseModel, Field")
        lines.append("from dataclasses import dataclass")
        lines.append("")

        custom_classes = []
        for typ in collected_types:
            # Check if it's a class with attributes
            if hasattr(typ, "__annotations__") or (
                hasattr(typ, "__dict__")
                and any(
                    not callable(getattr(typ, attr, None))
                    for attr in dir(typ)
                    if not attr.startswith("_")
                )
            ):
                custom_classes.append(typ)
        custom_classes = self._topological_sort_types(custom_classes, dependencies)

        # Generate custom classes (sorted by dependency)
        for cls in custom_classes:
            class_def = self._generate_class_definition(cls)
            if class_def:  # Only add non-empty class definitions
                lines.extend(class_def)
                lines.append("")

        if any(["Decimal" in line for line in lines]):
            lines.insert(2, "from decimal import Decimal")

        return "\n".join(lines)

    def _format_type(self, typ: type) -> str:
        if typ is None:
            return "Any"

        # Unwrap Annotated[T, ...]
        origin = get_origin(typ)
        if origin is Annotated:
            typ = get_args(typ)[0]
            origin = get_origin(typ)

        # Literal
        if origin is Literal:
            args = get_args(typ)
            literals = ", ".join(repr(a) for a in args)
            return f"Literal[{literals}]"

        # Union (Optional or other)
        if origin is Union:
            args = get_args(typ)
            non_none = [a for a in args if a is not type(None)]
            if len(non_none) == 1:
                return f"Optional[{self._format_type(non_none[0])}]"
            else:
                inner = ", ".join(self._format_type(a) for a in args)
                return f"Union[{inner}]"

        if origin is UnionType:
            args = get_args(typ)
            return "| ".join(self._format_type(a) for a in args)

        # Generic containers
        if origin:
            args = get_args(typ)
            inner = ", ".join(self._format_type(a) for a in args)
            if inner:
                return f"{_get_type_name(origin)}[{inner}]"
            return _get_type_name(origin)

        # Simple type
        return _get_type_name(origin or typ)


def _get_type_name(typ) -> str:
    """Get a consistent name for a type object."""
    if hasattr(typ, "__name__"):
        return typ.__name__
    return str(typ)


def _get_type_bases(typ: type) -> List[type]:
    if hasattr(typ, "__bases__"):
        return typ.__bases__  # type: ignore
    return []


def _is_global_or_class_function(func):
    if not callable(func):
        return False

    # Reject lambdas
    if _get_type_name(func) == "<lambda>":
        return False

    # Static methods and global functions are of type FunctionType
    if isinstance(func, types.FunctionType):
        return True

    # Class methods are MethodType but have __self__ as the class, not instance
    if isinstance(func, types.MethodType):
        if inspect.isclass(func.__self__):
            return True  # classmethod
        else:
            return False  # instance method

    return False



from pathlib import Path
from jinja2 import Environment, FileSystemLoader, select_autoescape

from ...common.py import path_to_module
from ...common.str import to_snake_case

TEMPLATES_DIR = Path(__file__).parent

env = Environment(
    loader=FileSystemLoader(str(TEMPLATES_DIR)),
    autoescape=select_autoescape(),
)
env.globals["path_to_module"] = path_to_module
env.globals["to_snake_case"] = to_snake_case


def load_template(template_name: str):
    return env.get_template(template_name)


from ...data_types import Domain, ToolGuardSpecItem
from mellea import generative


@generative
def tool_policy_pseudo_code(
    policy_item: ToolGuardSpecItem, fn_to_analyze: str, domain: Domain
) -> str:
    """
        Returns a pseudo code to check business constraints on a tool cool using an API

        Args:
            policy_item (ToolGuardSpecItem): Business policy, in natural language, specifying a constraint on a process involving the tool under analysis.
            fn_to_analyze (str): The function signature of the tool under analysis.
            domain (Domain): Python code defining available data types and APIs for invoking other tools.

        Returns:
            str: A pseudo code descibing how to use the API to check the tool call

        * The available API functions are listed in the `domain.app_api.content`.
        * Analyze the API functions' signatures (input and output parameter types).
        * You cannot assume other API functions.
        * For data objects (dataclasses or Pydantic models), only reference the explicitly declared fields.
            * Do not assume the presence of any additional fields.
            * Do not assume any implicit logic or relationships between field values (e.g., naming conventions).
        * List all the required API calls to check the business constraints.
        * If some information is missing, you should call another api function declared in the domain API.

        Examples:
    ```python
        domain = {
            "app_types": {
                "file_name": "car_types.py",
                "content": '''
                    class CarType(Enum):
                        SEDAN = "sedan"
                        SUV = "suv"
                        VAN = "van"
                    class Car:
                        plate_num: str
                        car_type: CarType
                    class Person:
                        id: str
                        driving_licence: str
                    class Insurance:
                        doc_id: str
                    class CarOwnership:
                        owenr_id: str
                        start_date: str
                        end_date: str
                '''
            },
            "app_api": {
                "file_name": "cars_api.py",
                "content": '''
                    class CarAPI(ABC):
                        def buy_car(self, plate_num: str, owner_id: str, insurance_id: str): pass
                        def get_person(self, id: str) -> Person: pass
                        def get_insurance(self, id: str) -> Insurance: pass
                        def get_car(self, plate_num: str) -> Car: pass
                        def car_ownership_history(self, plate_num: str) -> List[CarOwnership]: pass
                        def delete_car(self, plate_num: str): pass
                        def list_all_cars_owned_by(self, id: str) -> List[Car]: pass
                        def are_relatives(self, person1_id: str, person2_id: str) -> bool: pass
                '''
            }
        }
    ```
    * Example 1:
    ```
        tool_policy_pseudo_code(
            {"name": "documents exists", "description": "when buying a car, check that the car owner has a driving licence and that the insurance is valid."},
            "buy_car(plate_num: str, owner_id: str, insurance_id: str)",
            domain
        )
    ```
    may return:
    ```
        assert api.get_person(owner_id).driving_licence
        assert api.get_insurance(insurance_id)
    ```

    * Example 2:
    ```
        tool_policy_pseudo_code(
            {"name": "has driving licence", "description": "when buying a car, check that the car owner has a driving licence"},
            "buy_car(plate_num: str, owner_id: str, insurance_id: str)",
            domain
        )
    ```
        may return:
    ```
        assert api.get_insurance(insurance_id)
    ```

    * Example 3:
    ```
        tool_policy_pseudo_code(
            {"name": "no transfers on holidays", "description": "when buying a car, check that it is not a holiday today"},
            "buy_car(plate_num: str, owner_id: str, insurance_id: str)",
            domain
        )
    ```
        should return an empty string.

    * Example 4:
    ```
        tool_policy_pseudo_code(
            {"name": "Not in the same family",
            "description": "when buying a van, check that the van was never owned by someone from the buyer's family."},
            "buy_car(plate_num: str, owner_id: str, insurance_id: str)",
            domain
        )
    ```
        should return:
    ```
        car = api.get_car(plate_num)
        if car.car_type == CarType.VAN:
            history = api.car_ownership_history(plate_num)
            for each ownership in history:
                assert(not api.are_relatives(ownership.owenr_id, owner_id))
    ```
    """
    ...



from enum import Enum
from functools import cache
from typing import List


MEDIA_TYPE_APP_JSON = "application/json"
MEDIA_TYPE_MULTIPART_FORM = "multipart/form-data"
MEDIA_TYPE_APP_FORM = "application/x-www-form-urlencoded"


class StrEnum(str, Enum):
    """An abstract base class for string-based enums."""

    pass


class HttpMethod(StrEnum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"

    def __eq__(self, value: object) -> bool:
        return str(value).upper() == self.value

    def __ne__(self, value: object) -> bool:
        return not self.__eq__(value)

    def __hash__(self):
        return self.value.__hash__()

    @classmethod
    @cache
    def list(cls) -> List["HttpMethod"]:
        return list(map(lambda c: c.value, cls))


def is_valid_http_method(val: str):
    return val and val.upper() in HttpMethod.list()


PARAM_API_KEY = "api_key"

AUTH_HEADER = "Authorization"

SECURITY_COMPONENT_TYPE_API_KEY = "apiKey"
SECURITY_COMPONENT_SCHEME_BEARER = "bearer"
SECURITY_COMPONENT_SCHEME_BASIC = "basic"

SECURITY_COMPONENT_BEARER = {
    "type": "http",
    "scheme": SECURITY_COMPONENT_SCHEME_BEARER,
    "bearerFormat": "JWT",
}




import asyncio
import json
import os
from pathlib import Path
from typing import List, Optional

from ..data_types import ToolInfo, ToolGuardSpec, load_tool_policy
from ..llm.i_tg_llm import I_TG_LLM
from .utils import read_prompt_file, generate_messages, save_output, find_mismatched_references

class ToolGuardSpecGenerator:
    def __init__(self, llm:I_TG_LLM, policy_document:str, tools:List[ToolInfo], out_dir:str) -> None:
        self.llm = llm
        self.policy_document = policy_document
        self.tools_descriptions = {tool.name: tool.description for tool in tools}
        self.tools_details = {tool.name: tool for tool in tools}
        self.out_dir = out_dir
    

    async def generate_minimal_policy(self, tool_name: str) -> dict:
        tptd = await self.create_policy(tool_name)
        tptd = await self.example_creator(tool_name, tptd)
        return tptd

    async def generate_policy(self, tool_name: str) -> dict:
        tptd = await self.create_policy(tool_name)
        for i in range(3):
            tptd = await self.add_policies(tool_name, tptd, i)
        tptd = await self.split(tool_name, tptd)
        tptd = await self.merge(tool_name, tptd)
        tptd = await self.review_policy(tool_name, tptd)
        tptd = await self.add_references(tool_name, tptd)
        tptd = await self.reference_correctness(tool_name, tptd)
        tptd = await self.example_creator(tool_name, tptd)
        for i in range(5):  # FIXME
            tptd = await self.add_examples(tool_name, tptd, i)
        tptd = await self.merge_examples(tool_name, tptd)
        # tptd = self.fix_examples(tool_name, tptd)
        tptd = await self.review_examples(tool_name, tptd)
        return tptd

    async def create_policy(self, tool_name: str) -> dict:
        print("policy_creator_node")
        system_prompt = read_prompt_file("create_policy")
        system_prompt = system_prompt.replace("ToolX", tool_name)
        user_content = f"Policy Document:{self.policy_document}\nTools Descriptions:{json.dumps(self.tools_descriptions)}\nTarget Tool:{self.tools_details[tool_name].model_dump_json()}\n"
        tptd = await self.llm.chat_json(generate_messages(system_prompt, user_content))
        save_output(self.out_dir, f"{tool_name}.json", tptd)
        return tptd

    async def add_policies(
        self, tool_name: str, tptd: dict, iteration: int = 0
    ) -> dict:
        print("add_policy")
        system_prompt = read_prompt_file("add_policies")
        user_content = f"Policy Document:{self.policy_document}\nTools Descriptions:{json.dumps(self.tools_descriptions)}\nTarget Tool:{self.tools_details[tool_name].model_dump_json()}\nTPTD: {json.dumps(tptd)}"
        response = await self.llm.chat_json(
            generate_messages(system_prompt, user_content)
        )

        policies = (
            response["additionalProperties"]["policy_items"]
            if "additionalProperties" in response and "policy_items" not in response
            else response["policy_items"]
        )

        for policy in policies:
            # for policy in response["policy_items"]:
            policy["iteration"] = iteration
            tptd["policy_items"].append(policy)

        save_output(self.out_dir, f"{tool_name}_ADD_{iteration}.json", tptd)
        return tptd

    async def split(self, tool_name, tptd: dict) -> dict:
        # todo: consider addition step to split policy by policy and not overall
        print("split")
        system_prompt = read_prompt_file("split")
        user_content = f"Policy Document:{self.policy_document}\nTools Descriptions:{json.dumps(self.tools_descriptions)}\nTarget Tool:{self.tools_details[tool_name].model_dump_json()}\nTPTD: {json.dumps(tptd)}"
        tptd = await self.llm.chat_json(generate_messages(system_prompt, user_content))
        save_output(self.out_dir, f"{tool_name}_split.json", tptd)
        return tptd

    async def merge(self, tool_name, tptd: dict) -> dict:
        # todo: consider addition step to split policy by policy and not overall
        print("merge")
        system_prompt = read_prompt_file("merge")
        user_content = f"Policy Document:{self.policy_document}\nTools Descriptions:{json.dumps(self.tools_descriptions)}\nTarget Tool:{self.tools_details[tool_name].model_dump_json()}\nTPTD: {json.dumps(tptd)}"
        tptd = await self.llm.chat_json(generate_messages(system_prompt, user_content))

        save_output(self.out_dir, f"{tool_name}_merge.json", tptd)
        return tptd

    def move2archive(self, reviews) -> (bool, str):
        comments = ""
        num = len(reviews)
        if num == 0:
            return False
        counts = {
            "is_relevant": 0,
            "is_tool_specific": 0,
            "can_be_validated": 0,
            "is_actionable": 0,
        }

        for r in reviews:
            print(
                f"{r['is_relevant'] if 'is_relevant' in r else ''}\t{r['is_tool_specific'] if 'is_tool_specific' in r else ''}\t{r['can_be_validated'] if 'can_be_validated' in r else ''}\t{r['is_actionable'] if 'is_actionable' in r else ''}\t{r['is_self_contained'] if 'is_self_contained' in r else ''}\t{r['score'] if 'score' in r else ''}\t"
            )

            counts["is_relevant"] += r["is_relevant"] if "is_relevant" in r else 0
            counts["is_tool_specific"] += (
                r["is_tool_specific"] if "is_tool_specific" in r else 0
            )
            counts["can_be_validated"] += (
                r["can_be_validated"] if "can_be_validated" in r else 0
            )
            counts["is_actionable"] += r["is_actionable"] if "is_actionable" in r else 0

            if not all(
                e in r
                for e in [
                    "is_relevant",
                    "is_tool_specific",
                    "can_be_validated",
                    "is_actionable",
                ]
            ) or not (
                r["is_relevant"]
                and r["is_tool_specific"]
                and r["can_be_validated"]
                and r["is_actionable"]
            ):
                comments += r["comments"] + "\n"

        return not (all(float(counts[key]) / num > 0.5 for key in counts)), comments

    async def review_policy(self, tool_name, tptd) -> dict:
        print("review_policy")
        system_prompt = read_prompt_file("policy_reviewer")
        newTPTD = {"policy_items": []}

        if "policy_items" not in tptd:
            tptd["policy_items"] = []

        for item in tptd["policy_items"]:
            reviews = []
            for _iteration in range(5):
                user_content = f"Policy Document:{self.policy_document}\nTools Descriptions:{json.dumps(self.tools_descriptions)}\nTarget Tool:{json.dumps(self.tools_descriptions[tool_name])}\npolicy: {json.dumps(item)}"
                response = await self.llm.chat_json(
                    generate_messages(system_prompt, user_content)
                )
                if "is_self_contained" in response:
                    is_self_contained = response["is_self_contained"]
                    if not is_self_contained:
                        if "alternative_description" in response:
                            item["description"] = response["alternative_description"]
                        else:
                            print(
                                "Error: review is_self_contained is false but no alternative_description."
                            )
                else:
                    print("Error: review did not provide is_self_contained.")
                reviews.append(response)
            archive, comments = self.move2archive(reviews)
            print(archive)
            if archive:
                if "archive" not in newTPTD:
                    newTPTD["archive"] = []
                item["comments"] = comments
                newTPTD["archive"].append(item)
            else:
                newTPTD["policy_items"].append(item)
        save_output(self.out_dir, f"{tool_name}_rev.json", newTPTD)
        return newTPTD

    async def add_references(self, tool_name: str, tptd: dict) -> dict:
        print("add_ref")
        system_prompt = read_prompt_file("add_references")
        # remove old refs (used to help avoid duplications)
        for item in tptd["policy_items"]:
            item["references"] = []
            user_content = f"Policy Document:{self.policy_document}\nTools Descriptions:{json.dumps(self.tools_descriptions)}\nTarget Tool:{self.tools_details[tool_name].model_dump_json()}\npolicy: {json.dumps(item)}"
            response = await self.llm.chat_json(
                generate_messages(system_prompt, user_content)
            )
            if "references" in response:
                item["references"] = response["references"]
            else:
                print("Error! no references in response")
                print(response)

        save_output(self.out_dir, f"{tool_name}_ref.json", tptd)
        return tptd

    async def reference_correctness(self, tool_name: str, tptd: dict) -> dict:
        print("reference_correctness")
        tptd, unmatched_policies = find_mismatched_references(
            self.policy_document, tptd
        )
        save_output(self.out_dir, f"{tool_name}_ref_orig_.json", unmatched_policies)
        save_output(self.out_dir, f"{tool_name}_ref_correction_.json", tptd)
        return tptd

    async def example_creator(self, tool_name: str, tptd: dict) -> dict:
        print("example_creator")
        system_prompt = read_prompt_file("create_examples")
        system_prompt = system_prompt.replace("ToolX", tool_name)

        for item in tptd["policy_items"]:
            # user_content = f"Policy Document:{state['policy_text']}\nTools Descriptions:{json.dumps(state['tools'])}\nTarget Tool:{json.dumps(state['target_tool_description'])}\nPolicy:{policy}"
            user_content = f"Tools Descriptions:{json.dumps(self.tools_descriptions)}\nTarget Tool:{self.tools_details[tool_name].model_dump_json()}\nPolicy:{item}"

            response = await self.llm.chat_json(
                generate_messages(system_prompt, user_content)
            )
            if "violation_examples" in response:
                item["violation_examples"] = response["violation_examples"]

            if "compliance_examples" in response:
                item["compliance_examples"] = response["compliance_examples"]

        save_output(self.out_dir, f"{tool_name}_examples.json", tptd)
        return tptd

    async def add_examples(self, tool_name: str, tptd: dict, iteration: int) -> dict:
        print("add_examples")
        system_prompt = read_prompt_file("add_examples")
        system_prompt = system_prompt.replace("ToolX", tool_name)
        for item in tptd["policy_items"]:
            # user_content = f"Policy Document:{state['policy_text']}\nTools Descriptions:{json.dumps(state['tools'])}\nTarget Tool:{json.dumps(state['target_tool_description'])}\nPolicy:{policy}"
            user_content = f"Tools Descriptions:{json.dumps(self.tools_descriptions)}\nTarget Tool:{self.tools_details[tool_name].model_dump_json()}\nPolicy:{item}"
            response = await self.llm.chat_json(
                generate_messages(system_prompt, user_content)
            )
            if "violation_examples" in response:
                for vexample in response["violation_examples"]:
                    # vexample["iteration"] = state["iteration"]
                    if "violation_examples" not in item:
                        item["violation_examples"] = []
                    item["violation_examples"].append(vexample)
            if "compliance_examples" in response:
                for cexample in response["compliance_examples"]:
                    if "compliance_examples" not in item:
                        item["compliance_examples"] = []
                    # cexample["iteration"] = state["iteration"]
                    item["compliance_examples"].append(cexample)

        save_output(self.out_dir, f"{tool_name}_ADD_examples{iteration}.json", tptd)
        return tptd

    async def merge_examples(self, tool_name: str, tptd: dict) -> dict:
        print("merge_examples")
        system_prompt = read_prompt_file("merge_examples")
        system_prompt = system_prompt.replace("ToolX", tool_name)
        for item in tptd["policy_items"]:
            # user_content = f"Policy Document:{state['policy_text']}\nTools Descriptions:{json.dumps(state['tools'])}\nTarget Tool:{json.dumps(state['target_tool_description'])}\nPolicy Name:{policy['name']}\nPolicy Description:{policy['description']}"
            user_content = f"Tools Descriptions:{json.dumps(self.tools_descriptions)}\nTarget Tool:{self.tools_details[tool_name].model_dump_json()}\nPolicy Name:{item['name']}\nPolicy Description:{item['description']}"
            user_content += f"\n\nviolation Examples: {item['violation_examples']}"
            user_content += f"\n\nCompliance Examples: {item['compliance_examples']}"
            response = await self.llm.chat_json(
                generate_messages(system_prompt, user_content)
            )
            item["violation_examples"] = response["violation_examples"]
            item["compliance_examples"] = response["compliance_examples"]

        save_output(self.out_dir, f"{tool_name}_merge_examples.json", tptd)
        return tptd

    async def fix_examples(self, tool_name: str, tptd: dict) -> dict:
        print("fix_examples")
        orig_prompt = read_prompt_file("fix_example")
        for item in tptd["policy_items"]:
            for etype in ["violation", "compliance"]:
                fixed_examples = []
                for example in item[etype + "_examples"]:
                    system_prompt = orig_prompt.replace("ToolX", tool_name)
                    system_prompt = system_prompt.replace("__EXAMPLE_TYPE__", "")

                    # user_content = f"Policy Document:{state['policy_text']}\nTools Descriptions:{json.dumps(state['tools'])}\nTarget Tool:{json.dumps(state['target_tool_description'])}\nPolicy Name:{policy['name']}\nPolicy Description:{policy['description']}\nExample:{example}"
                    user_content = f"Tools Descriptions:{json.dumps(self.tools_descriptions)}\nTarget Tool:{self.tools_details[tool_name].model_dump_json()}\nPolicy Name:{item['name']}\nPolicy Description:{item['description']}\nExample:{example}"

                    response = await self.llm.chat_json(
                        generate_messages(system_prompt, user_content)
                    )
                    fixed_examples.append(response["revised_example"])
                item[etype + "_examples"] = fixed_examples

        save_output(self.out_dir, f"{tool_name}_fix_examples.json", tptd)
        return tptd

    # todo: change to revew examples, write prompts
    async def review_examples(self, tool_name: str, tptd: dict) -> dict:
        print("review_examples")
        system_prompt = read_prompt_file("examples_reviewer")
        for item in tptd["policy_items"]:
            print(item["name"])
            for etype in ["violation", "compliance"]:
                print(etype)
                passed_examples = []
                for example in item[etype + "_examples"]:
                    print(example)
                    reviews = []
                    for _iteration in range(5):
                        # user_content = f"Policy Document:{state['policy_text']}\nTools Descriptions:{json.dumps(state['tools'])}\nTarget Tool:{json.dumps(state['target_tool_description'])}\nPolicy Name:{policy['name']}\nPolicy Description:{policy['description']}\nExample:{example}"
                        user_content = f"Tools Descriptions:{json.dumps(self.tools_descriptions)}\nTarget Tool:{self.tools_details[tool_name].model_dump_json()}\nPolicy Name:{item['name']}\nPolicy Description:{item['description']}\nExample:{example}"
                        response = await self.llm.chat_json(
                            generate_messages(system_prompt, user_content)
                        )
                        reviews.append(response)
                    keep = self.keep_example(reviews)
                    if keep:
                        passed_examples.append(example)

                item[etype + "_examples"] = passed_examples

        save_output(self.out_dir, f"{tool_name}_example_rev.json", tptd)
        return tptd

    def keep_example(self, reviews) -> bool:
        bads = 0
        totals = 0
        for r in reviews:
            for vals in r.values():
                totals += 1
                if "value" not in vals:
                    print(reviews)
                elif not vals["value"]:
                    bads += 1
        if bads / totals > 0.8:
            return False
        return True


async def extract_toolguard_specs(
    policy_text: str,
    tools: List[ToolInfo],
    step1_output_dir: str,
    llm: I_TG_LLM,
    tools_shortlist: Optional[List[str]] = None,
    short=False,
) -> List[ToolGuardSpec]:
    if not os.path.isdir(step1_output_dir):
        os.makedirs(step1_output_dir)

    process_dir = os.path.join(step1_output_dir, "process")
    if not os.path.isdir(process_dir):
        os.makedirs(process_dir)
    tpg = ToolGuardSpecGenerator(llm, policy_text, tools, process_dir)

    async def do_one_tool(tool_name)->ToolGuardSpec:
        spec_dict = await tpg.generate_minimal_policy(tool_name) if short \
            else await tpg.generate_policy(tool_name)
        spec = ToolGuardSpec(tool_name=tool_name, **spec_dict)

        path = Path(step1_output_dir, tool_name + ".json")
        path.write_text(
            spec.model_dump_json(indent=2),
            encoding="utf-8"
        )
        return spec

    specs = await asyncio.gather(
        *[
            do_one_tool(tool.name)
            for tool in tools
            if ((tools_shortlist is None) or (tool.name in tools_shortlist))
        ]
    )
    print("All tools done")
    return specs


from typing import List
from ...data_types import Domain, ToolGuardSpecItem
from mellea import generative

# from toolguard.gen_py.prompts.python_code import PythonCodeModel


@generative
def improve_tool_guard(
    prev_impl: str,
    domain: Domain,
    policy_item: ToolGuardSpecItem,
    dependent_tool_names: List[str],
    review_comments: List[str],
) -> str:
    """
        Improve the previous tool-call guard implementation (in Python) so that it fully adheres to the given policy and addresses all review comments.

        Args:
            prev_impl (str): The previous implementation of the tool-call check.
            domain (Domain): Python code defining available data types and other tool interfaces.
            policy_item (ToolGuardSpecItem): Requirements for this tool.
            dependent_tool_names (List[str]): Names of other tools that this tool may call to obtain required information.
            review_comments (List[str]): Review feedback on the current implementation (e.g., pylint errors, failed unit tests).

        Returns:
            str: The improved implementation of the tool-call check.

        Implementation Rules:
            - Do not modify the function signature, parameter names, or type annotations.
            - All policy requirements must be validated.
            - Keep the implementation simple and well-documented.
            - Only validate the tool-call arguments; never call the tool itself.
            - If additional information is needed beyond the function arguments, use only the APIs of tools listed in `dependent_tool_names`.
            - Generate code that enforces the given policy only, do not generate any additional logic that is not explicitly mentioned in the policy.

        **Example: **
    prev_impl = ```python
    from typing import *
    from airline.airline_types import *
    from airline.i_airline import I_Airline

    def guard_Checked_Bag_Allowance_by_Membership_Tier(api: I_Airline, user_id: str, passengers: list[Passenger]):
        \"\"\"
        Limit to five passengers per reservation.
        \"\"\"
        pass #FIXME
    ```

    should return something like:
    ```python
    from typing import *
    from airline.airline_types import *
    from airline.i_airline import I_Airline

    def guard_Checked_Bag_Allowance_by_Membership_Tier(api: I_Airline, user_id: str, passengers: list[Passenger]):
        \"\"\"
        Limit to five passengers per reservation.
        \"\"\"
        if len(passengers) > 5:
            raise PolicyViolationException("More than five passengers are not allowed.")
    ```
    """
    ...



import asyncio
import logging
from typing import Callable, List, Optional

import mellea

from .mellea_simple import SimpleBackend
from ..llm.i_tg_llm import I_TG_LLM
from .domain_from_funcs import generate_domain_from_functions
from ..data_types import RuntimeDomain, ToolGuardSpec
from .domain_from_openapi import generate_domain_from_openapi
from ..runtime import ToolGuardsCodeGenerationResult
from .tool_guard_generator import ToolGuardGenerator
from .utils import pytest, pyright

logger = logging.getLogger(__name__)

async def generate_toolguards_from_functions(
    app_name: str,
    tool_policies: List[ToolGuardSpec],
    py_root: str,
    funcs: List[Callable],
    llm: I_TG_LLM,
    module_roots: Optional[List[str]] = None,
) -> ToolGuardsCodeGenerationResult:
    assert funcs, "Funcs cannot be empty"
    logger.debug(f"Starting... will save into {py_root}")

    if not module_roots:
        if len(funcs) > 0:
            module_roots = list(
                {func.__module__.split(".")[0] for func in funcs}
            )
    assert module_roots

    # Domain from functions
    domain = generate_domain_from_functions(py_root, app_name, funcs, module_roots)
    return await generate_toolguards_from_domain(
        app_name, tool_policies, py_root, domain, llm
    )


async def generate_toolguards_from_openapi(
    app_name: str, tool_policies: List[ToolGuardSpec], py_root: str, openapi_file: str, llm: I_TG_LLM
) -> ToolGuardsCodeGenerationResult:
    logger.debug(f"Starting... will save into {py_root}")

    # Domain from OpenAPI
    domain = generate_domain_from_openapi(py_root, app_name, openapi_file)
    return await generate_toolguards_from_domain(
        app_name, tool_policies, py_root, domain, llm
    )

async def generate_toolguards_from_domain(
    app_name: str, specs: List[ToolGuardSpec], py_root: str, domain: RuntimeDomain,
    llm: I_TG_LLM
) -> ToolGuardsCodeGenerationResult:
    # Setup env
    pyright.config(py_root)
    pytest.configure(py_root)

    for tool_policy in specs:
        for policy in tool_policy.policy_items:
            policy.name = policy.name.replace(".","_")

    not_empty_specs = [spec for spec in [
        ToolGuardSpec( #a copy
            tool_name=spec.tool_name, 
            policy_items=[i for i in spec.policy_items if not i.skip])
        for spec in specs
    ] if len(spec.policy_items) > 0]

    # mellea_workaround = {"model_options": {"reasoning_effort": "medium"}}#FIXME https://github.com/generative-computing/mellea/issues/270
    # kw_args = llm.kw_args
    # kw_args.update(mellea_workaround)
    mellea_backend = SimpleBackend(llm)
    m = mellea.MelleaSession(mellea_backend)
    tool_results = await asyncio.gather(
        *[
            ToolGuardGenerator(
                app_name, tool_policy, py_root, domain, m
            ).generate()
            for tool_policy in not_empty_specs
        ]
    )

    tools_result = {
        tool.tool_name: res for tool, res in zip(not_empty_specs, tool_results)
    }
    return ToolGuardsCodeGenerationResult(
        out_dir=py_root, 
        domain=domain, 
        tools=tools_result
    ).save(py_root)



import re


def to_camel_case(snake_str: str) -> str:
    return (
        snake_str.replace("_", " ")
        .title()
        .replace(" ", "")
        .replace("-", "_")
        .replace("'", "_")
        .replace(",", "_")
        .replace("’","_")
    )


def to_snake_case(human_name: str) -> str:
    return (
        human_name.lower()
        .replace(" ", "_")
        .replace("-", "_")
        .replace("'", "_")
        .replace(",", "_")
        .replace("’","_")
    )
      
def to_pascal_case(name: str) -> str:
    # Split by underscores first
    parts = name.split('_')

    result_parts = []
    for part in parts:
        # Split camelCase or mixedCase segments into words
        subparts = re.findall(r'[A-Z]?[a-z]+|[A-Z]+(?![a-z])|\d+', part)
        # Capitalize each segment
        cap = ''.join(s[0].upper() + s[1:] if s else '' for s in subparts)
        result_parts.append(cap)

    return ''.join(result_parts)


import asyncio
import json
import os
import re
from typing import Any, List, Dict
import time
from litellm import completion
from litellm.exceptions import RateLimitError
import json
from abc import ABC
import dotenv

from ..llm.i_tg_llm import I_TG_LLM


class LanguageModelBase(I_TG_LLM, ABC):

	def __init__(self, model_name:str) -> None:
		self.model_name = model_name

	async def chat_json(self, messages: List[Dict], max_retries: int = 5, backoff_factor: float = 1.5) -> Dict:
		retries = 0
		while retries < max_retries:
			try:
				response = await self.generate(messages)
				res = self.extract_json_from_string(response)
				if res is None:
					wait_time = backoff_factor ** retries
					print(f"Error: not json format. Retrying in {wait_time:.1f} seconds... (attempt {retries + 1}/{max_retries})")
					time.sleep(wait_time)
					retries += 1
				else:
					return res
			except RateLimitError as e:
				wait_time = backoff_factor ** retries
				print(f"Rate limit hit. Retrying in {wait_time:.1f} seconds... (attempt {retries + 1}/{max_retries})")
				time.sleep(wait_time)
				retries += 1
			except Exception as e:
				raise RuntimeError(f"Unexpected error during chat completion: {e}")
		raise RuntimeError("Exceeded maximum retries due to rate limits.")
	
	def extract_json_from_string(self,s):
		# Use regex to extract the JSON part from the string
		match = re.search(r'```json\s*(\{.*?\})\s*```', s, re.DOTALL)
		if match:
			json_str = match.group(1)
			try:
				return json.loads(json_str)
			except json.JSONDecodeError as e:
				print(f"Failed to decode JSON: {e}")
				return None
		else:
			## for rits?
			match = re.search(r'(\{[\s\S]*\})', s)
			if match:
				json_str = match.group(1)
				try:
					return json.loads(json_str)
				except json.JSONDecodeError as e:
					print("response:")
					print(s)
					print("match:")
					print(match.group(1))
					print("Failed to parse JSON:", e)
					return None
			
			print("No JSON found in the string.")
			print(s)
			return None

			
rits_model_name_to_endpoint_list=[
#	{"endpoint":"https://ete-litellm.bx.cloud9.ibm.com", "model_name":"claude-3-7-sonnet"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/avengers-jamba-9b","model_name":"ibm-fms/avengers-jamba-9b"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/codellama-34b-instruct-hf","model_name":"codellama/CodeLlama-34b-Instruct-hf"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/deepseek-coder-33b-instruct","model_name":"deepseek-ai/DeepSeek-V2.5"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/deepseek-v2-5/v1","model_name":"deepseek-ai/DeepSeek-V2.5"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/deepseek-v3/v1","model_name":"deepseek-ai/DeepSeek-V3"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/gemma-2-9b-it","model_name":"google/gemma-2-9b-it"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-13b-chat-v2","model_name":"ibm-granite/granite-13b-chat-v2"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-13b-instruct-v2","model_name":"ibm-granite/granite-13b-instruct-v2"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-20b-cbcl","model_name":"ibm-granite/granite-20b-code-base-content-linking"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-20b-cbsl","model_name":"ibm/granite-20b-schema-sqlinstruct-granite-fine-vs"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-20b-cbsql","model_name":"ibm-granite/granite-20b-code-base-sql-gen"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-20b-ci-r1-1","model_name":"ibm-granite/granite-20b-code-instruct-r1.1"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-20b-code-8k-ansible","model_name":"ibm/granite-20b-code-8k-ansible"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-20b-code-instruct","model_name":"ibm/granite-20b-code-instruct"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-20b-code-instruct-8k","model_name":"ibm-granite/granite-20b-code-instruct-8k"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-20b-code-instruct-uapi","model_name":"ibm-granite/granite-20b-code-instruct-unified-api"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-20b-functioncalling","model_name":"ibm-granite/granite-20b-functioncalling"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-3-0-8b-instruct","model_name":"ibm-granite/granite-3.0-8b-instruct"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-3-1-8b-content-linking","model_name":"ibm/granite-3-1-8b-content-linking"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-3-1-8b-instruct","model_name":"ibm-granite/granite-3.1-8b-instruct"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-3-1-8b-question-gen","model_name":"ibm/granite-3-1-8b-question-gen"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-3-1-8b-schema-linking","model_name":"ibm/granite-3-1-8b-schema-linking"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-3-1-8b-sql-gen","model_name":"ibm/granite-3-1-8b-sql-gen"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-3-2-8b-instruct","model_name":"ibm-granite/granite-3.2-8b-instruct"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-3-3-8b-instruct/v1/","model_name":"ibm-granite/granite-3.3-8b-instruct"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-34b-code-instruct-8k","model_name":"ibm-granite/granite-34b-code-instruct-8k"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-34b-content-linking","model_name":"ibm/granite-34b-content-linking"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-34b-question-gen","model_name":"ibm/granite-34b-question-gen"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-34b-schema-linking","model_name":"ibm/granite-34b-schema-linking"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-34b-sql-gen","model_name":"ibm/granite-34b-sql-gen"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-7b-lab","model_name":"ibm/granite-7b-lab"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-8b-code-instruct-128k","model_name":"ibm-granite/granite-8b-code-instruct-128k"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-8b-code-instruct-4k","model_name":"ibm-granite/granite-8b-code-instruct-4k"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-8b-instruct-preview-4k","model_name":"ibm-granite/granite-8b-instruct-preview-4k"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-8b-japanese-instruct","model_name":"ibm-granite/granite-8b-japanese-instruct"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/granite-vision-3-2-2b","model_name":"ibm-granite/granite-vision-3.2-2b"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/internvl2-llama3-76b","model_name":"OpenGVLab/InternVL2-Llama3-76B"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/llama-3-1-405b-instruct-fp8","model_name":"meta-llama/llama-3-1-405b-instruct-fp8"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/llama-3-1-70b-instruct","model_name":"meta-llama/llama-3-1-70b-instruct"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/llama-3-1-8b-instruct","model_name":"meta-llama/Llama-3.1-8B-Instruct"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/llama-3-1-test","model_name":"meta-llama/llama-3-1-405b-instruct-fp8"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/llama-3-2-11b-instruct","model_name":"meta-llama/Llama-3.2-11B-Vision-Instruct"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/llama-3-2-90b-instruct","model_name":"meta-llama/Llama-3.2-90B-Vision-Instruct"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/llama-3-3-70b-instruct/v1","model_name":"meta-llama/llama-3-3-70b-instruct"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/microsoft-phi-4/v1","model_name":"microsoft/phi-4"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/mistralai-pixtral-12b-2409","model_name":"mistralai/pixtral-12b-2409"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/mixtral-8x22b-instruct-v01/v1","model_name":"mistralai/mixtral-8x22B-instruct-v0.1"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/mixtral-8x7b-instruct-v01/v1","model_name":"mistralai/mixtral-8x7B-instruct-v0.1"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/mixtral-8x7b-instruct-v01-test","model_name":"mistralai/mixtral-8x7B-instruct-v0.1-test"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/pixtral-large-instruct-2411","model_name":"mistralai/Pixtral-Large-Instruct-2411"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/qwen2-5-72b-instruct/v1","model_name":"Qwen/Qwen2.5-72B-Instruct"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/qwen2-vl-72b-instruct","model_name":"Qwen/Qwen2-VL-72B-Instruct"},
	{"endpoint":"https://inference-3scale-apicast-production.apps.rits.fmaas.res.ibm.com/slate-125m-english-rtrvr-v2","model_name":"ibm/slate-125m-english-rtrvr-v2"}]

rits_model_to_endpoint = {
    entry["model_name"]: entry["endpoint"]
    for entry in rits_model_name_to_endpoint_list
}

anthropic_models = ['claude-3-5-sonnet-latest', 'claude-3-5-haiku']
RITS = "RITS"

		
class LitellmModel(LanguageModelBase):
	def __init__(self, model_name:str, provider: str, kw_args: Dict[str, Any] = {}):
		super().__init__(model_name)
		self.provider = provider
		self.kw_args = kw_args

	async def generate(self, messages: List[Dict])->str:
		provider = self.provider
		base_url = None
		extra_headers = {
			'Content-Type': 'application/json'
		}
		if self.provider and self.provider.upper() == RITS:
			provider = "openai"
			base_url = rits_model_to_endpoint[self.model_name]
			extra_headers['RITS_API_KEY'] = os.getenv("RITS_API_KEY") or ''
		
		response = completion(
			messages=messages,
			model=self.model_name,
			custom_llm_provider=provider,
			base_url=base_url,
			extra_headers=extra_headers,
			**self.kw_args
		)
		return response.choices[0].message.content
	
	
if __name__ == '__main__':
	dotenv.load_dotenv()
	model = "gpt-4o-2024-08-06"
	#model = "claude-3-5-sonnet-20240620"
	#model = "meta-llama/llama-3-3-70b-instruct"
	aw = LitellmModel(model, "azure")

	async def my_test():
		resp = await aw.generate([{"role": "user", "content": "what is the weather?"}])
		print(resp)
		resp = await aw.chat_json([{"role": "user", "content": "what is the weather? please answer in json format"}])
		print(resp)
	asyncio.run(my_test())
	




import asyncio
import inspect
import os
from os.path import join
from typing import Callable, List

import markdown
import logging

from .buildtime import generate_guard_specs, generate_guards_from_specs
from .common import py
from .llm.tg_litellm import LitellmModel

logger = logging.getLogger(__name__)

import argparse

def main():
	parser = argparse.ArgumentParser(description='parser')
	parser.add_argument('--policy-path', type=str, help='Path to the policy file. Currently, in `markdown` syntax. eg: `/Users/me/airline/wiki.md`')
	parser.add_argument('--app-root', type=str, default="../ToolGuardAgent/src/appointment_app")
	parser.add_argument('--app-tools-py-file', type=str, default="lg_tools.py")
	parser.add_argument('--out-dir', type=str, help='Path to an output folder where the generated artifacts will be written. eg: `/Users/me/airline/outdir2')
	parser.add_argument('--step1-dir-name', type=str, default='Step1', help='Step1 folder name under the output folder')
	parser.add_argument('--step2-dir-name', type=str, default='Step2', help='Step2 folder name under the output folder')
	parser.add_argument('--step1-model-name', type=str, default='gpt-4o-2024-08-06', help='Model to use for generating in step 1')
	parser.add_argument('--tools2run', nargs='+', default=None, help='Optional list of tool names. These are a subset of the tools in the openAPI operation ids.')
	parser.add_argument('--short-step1', action='store_true', default=False, help='run short version of step 1')
	
	args = parser.parse_args()
	policy_path = args.policy_path
	
	policy_text = open(policy_path, 'r', encoding='utf-8').read()
	policy_text = markdown.markdown(policy_text)

	llm = LitellmModel(args.step1_model_name, "azure") #FIXME from args
	tools = load_functions_in_file(args.app_root, args.app_tools_py_file)
	os.makedirs(args.out_dir, exist_ok=True)
	step1_out_dir = join(args.out_dir, args.step1_dir_name)
	step2_out_dir = join(args.out_dir, args.step2_dir_name)

	async def build_toolguards():
		specs = await generate_guard_specs(
			policy_text = policy_text, 
			tools = tools, 
			work_dir = step1_out_dir, 
			llm = llm,
			short=args.short_step1,
		)
		guards = await generate_guards_from_specs(
			tool_specs = specs,
			tools = tools, 
			work_dir = step2_out_dir, 
			llm = llm,
			tool_names=args.tools2run
		)
		return guards
		
	asyncio.run(
		build_toolguards()
	)

from importlib import import_module
def load_functions_in_file(py_root:str, file_path: str) -> List[Callable]:
	with py.temp_python_path(py_root):
		module = import_module(py.path_to_module(file_path))
	funcs = []
	for name, obj in inspect.getmembers(module):
		# if isinstance(obj, BaseTool):
		# 	funcs.append(py.unwrap_fn(obj))
		if callable(obj) and not (name=='tool' and obj.__module__ =='langchain_core.tools.convert'):
			funcs.append(obj)
	return funcs


import functools
from typing import Callable, List, TypeVar, Any


def flatten(arr_arr):
    return [b for bs in arr_arr for b in bs]


def break_array_into_chunks(arr: List, chunk_size: int) -> List[List[Any]]:
    res = []
    for i, v in enumerate(arr):
        if i % chunk_size == 0:
            cur = []
            res.append(cur)
        cur.append(v)
    return res


def sum(array):
    return functools.reduce(lambda a, b: a + b, array) if len(array) > 0 else 0


T = TypeVar("T")


def find(array: List[T], pred: Callable[[T], bool]):
    for item in array:
        if pred(item):
            return item


# remove duplicates and preserve ordering
T = TypeVar("T")


def remove_duplicates(array: List[T]) -> List[T]:
    res = []
    visited = set()
    for item in array:
        if item not in visited:
            res.append(item)
            visited.add(item)
    return res


def not_none(array: List[T]) -> List[T]:
    return [item for item in array if item is not None]


def split_array(arr: List[T], delimiter: T) -> List[List[T]]:
    result = []
    temp = []
    for item in arr:
        if item == delimiter:
            if temp:  # Avoid adding empty lists
                result.append(temp)
            temp = []
        else:
            temp.append(item)
    if temp:  # Add the last group if not empty
        result.append(temp)
    return result



from .buildtime import generate_guard_specs, generate_guards_from_specs
from .llm.tg_litellm import LitellmModel, I_TG_LLM
from .data_types import *

from .runtime import IToolInvoker, ToolFunctionsInvoker, ToolGuardsCodeGenerationResult, ToolMethodsInvoker, load_toolguard_code_result, load_toolguards, LangchainToolInvoker


from typing import Dict, Any, Callable
import hashlib
import json
from collections import deque


def dict_deep_merge(trg_dct: Dict, merge_dct: Dict):
    for k, v in merge_dct.items():
        trg_v = trg_dct.get(k)
        if k in trg_dct and isinstance(trg_v, dict) and isinstance(v, dict):
            dict_deep_merge(trg_v, v)
        elif isinstance(trg_v, list) and isinstance(v, list):
            if all([type(item) is dict for item in v]) and all(
                [type(item) is dict for item in trg_v]
            ):  # both are lists of objects
                for a, b in zip(trg_v, v):
                    dict_deep_merge(a, b)
            else:
                for item in v:
                    if item not in trg_dct[k]:
                        trg_dct[k].append(item)
        else:
            trg_dct[k] = v


def get_keys_recursive(d: Dict):
    keys = []
    for key, value in d.items():
        keys.append(key)
        if isinstance(value, dict):
            keys.extend(get_keys_recursive(value))
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    keys.extend(get_keys_recursive(item))
    return keys


# adapted from https://www.doc.ic.ac.uk/~nuric/posts/coding/how-to-hash-a-dictionary-in-python/
def dict_hash(dictionary: Dict[str, Any]) -> int:
    encoded = json.dumps(dictionary, sort_keys=True).encode()
    return int(hashlib.sha256(encoded).hexdigest(), 16)


def visit_all(d: Any, cb: Callable[[Dict, Any], bool]):
    if type(d) is dict:
        for k in list(
            d
        ):  # list() to avoid RuntimeError: dictionary changed size during iteration
            changed = cb(d, k)
            if not changed:
                visit_all(d[k], cb)
    if type(d) is list:
        for item in d:
            visit_all(item, cb)


def resolve_ref(ref, root_schema, resolved_refs):
    """Resolve a $ref to its actual definition in the schema."""
    if ref in resolved_refs:
        return resolved_refs[ref]

    resolved_schema = find_ref(root_schema, ref)
    resolved_refs[ref] = resolved_schema
    return substitute_refs(resolved_schema, root_schema, resolved_refs)


def substitute_refs(schema, root_schema=None, resolved_refs=None):
    """Substitute all $refs in the JSON schema with their definitions."""
    if root_schema is None:
        root_schema = schema
    if resolved_refs is None:
        resolved_refs = {}

    if isinstance(schema, dict):
        if "$ref" in schema:
            ref = schema["$ref"]
            if ref in resolved_refs:
                return resolved_refs[ref]
            resolved = resolve_ref(ref, root_schema, resolved_refs)
            resolved_refs[ref] = resolved
            return resolved
        else:
            return {
                k: substitute_refs(v, root_schema, resolved_refs)
                for k, v in schema.items()
            }
    elif isinstance(schema, list):
        return [substitute_refs(item, root_schema, resolved_refs) for item in schema]
    else:
        return schema


def find_ref(doc: dict, ref: str):
    q = deque(ref.split("/"))
    cur = doc
    while q:
        field = q.popleft()
        if field == "#":
            cur = doc
            continue
        if field in cur:
            cur = cur[field]
        else:
            return None
    if "$ref" in cur:
        return find_ref(doc, cur.get("$ref"))  # recursive. infinte loops?
    return cur



import os
from pathlib import Path
from typing import Callable, List
from os.path import join

from .api_extractor import APIExtractor
from ..common.str import to_camel_case, to_snake_case
from ..data_types import FileTwin, RuntimeDomain
from . import consts


def generate_domain_from_functions(
    py_path: str, app_name: str, funcs: List[Callable], include_module_roots: List[str]
) -> RuntimeDomain:
    # ToolGuard Runtime
    os.makedirs(join(py_path, consts.RUNTIME_PACKAGE_NAME), exist_ok=True)

    root = str(Path(__file__).parent.parent)
    common = FileTwin.load_from(root, "data_types.py").save_as(
        py_path, join(consts.RUNTIME_PACKAGE_NAME, consts.RUNTIME_TYPES_PY)
    )
    runtime = FileTwin.load_from(root, "runtime.py")
    # runtime.content = runtime.content.replace(
    #     "toolguard.", f"{consts.RUNTIME_PACKAGE_NAME}."
    # )
    runtime.save_as(py_path, join(consts.RUNTIME_PACKAGE_NAME, consts.RUNTIME_INIT_PY))

    # APP init and Types
    os.makedirs(join(py_path, to_snake_case(app_name)), exist_ok=True)
    FileTwin(file_name=join(to_snake_case(app_name), "__init__.py"), content="").save(
        py_path
    )

    extractor = APIExtractor(py_path=py_path, include_module_roots=include_module_roots)
    api_cls_name = f"I_{to_camel_case(app_name)}"
    impl_module_name = to_snake_case(f"{app_name}.{app_name}_impl")
    impl_class_name = to_camel_case(f"{app_name}_Impl")
    api, types, impl = extractor.extract_from_functions(
        funcs,
        interface_name=api_cls_name,
        interface_module_name=to_snake_case(f"{app_name}.i_{app_name}"),
        types_module_name=to_snake_case(f"{app_name}.{app_name}_types"),
        impl_module_name=impl_module_name,
        impl_class_name=impl_class_name,
    )

    return RuntimeDomain(
        app_name=app_name,
        toolguard_common=common,
        app_types=types,
        app_api_class_name=api_cls_name,
        app_api=api,
        app_api_impl_class_name=impl_class_name,
        app_api_impl=impl,
        app_api_size=len(funcs),
    )



from enum import Enum
from typing import Any, Dict, List, Optional, Union


from .ref import DocumentWithRef, Reference


class StrEnum(str, Enum):
    """An abstract base class for string-based enums."""
    pass


class JSONSchemaTypes(StrEnum):
    string = "string"
    number = "number"
    integer = "integer"
    boolean = "boolean"
    array = "array"
    object = "object"
    null = "null"


class JSchema(DocumentWithRef):
    type: Optional[JSONSchemaTypes] = None
    properties: Optional[Dict[str, Union[Reference, "JSchema"]]] = None
    items: Optional[Union[Reference, "JSchema"]] = None
    additionalProperties: Optional[Union["JSchema", bool]] = None
    format: Optional[str] = None
    enum: Optional[list] = None
    default: Optional[Any] = None
    description: Optional[str] = None
    example: Optional[Any] = None
    required: Optional[List[str]] = None
    allOf: Optional[List[Union[Reference, "JSchema"]]] = None
    anyOf: Optional[List[Union[Reference, "JSchema"]]] = None
    nullable: Optional[bool] = (
        None  # in OPenAPISpec https://swagger.io/docs/specification/v3_0/data-models/data-types/#null
    )

    def __str__(self) -> str:
        return self.model_dump_json(exclude_none=True, indent=2)



import os
from pathlib import Path
from typing import List, Optional, Tuple, Union
from os.path import join

from ..common.array import find
from ..common.py import module_to_path
from ..common.str import to_camel_case, to_pascal_case, to_snake_case
from . import consts
from .templates import load_template
from .utils.datamodel_codegen import (
    run as dm_codegen,
)
from ..common.open_api import (
    OpenAPI,
    Operation,
    Parameter,
    ParameterIn,
    PathItem,
    Reference,
    RequestBody,
    Response,
    JSchema,
    read_openapi,
)
from ..data_types import FileTwin, RuntimeDomain

ARGS = "args"

def generate_domain_from_openapi(
    py_path: str, app_name: str, openapi_file: str
) -> RuntimeDomain:
    # ToolGuard Runtime
    os.makedirs(join(py_path, consts.RUNTIME_PACKAGE_NAME), exist_ok=True)

    root = str(Path(__file__).parent.parent)
    common = FileTwin.load_from(root, "data_types.py").save_as(
        py_path, join(consts.RUNTIME_PACKAGE_NAME, consts.RUNTIME_TYPES_PY)
    )
    runtime = FileTwin.load_from(root, "runtime.py")
    # runtime.content = runtime.content.replace(
    #     "toolguard.", f"{consts.RUNTIME_PACKAGE_NAME}."
    # )
    runtime.save_as(py_path, join(consts.RUNTIME_PACKAGE_NAME, consts.RUNTIME_INIT_PY))

    # APP Types
    oas = read_openapi(openapi_file)
    os.makedirs(join(py_path, to_snake_case(app_name)), exist_ok=True)

    types_name = f"{app_name}_types"
    types_module_name = f"{app_name}.{types_name}"
    types = FileTwin(
        file_name=module_to_path(types_module_name), content=dm_codegen(openapi_file)
    ).save(py_path)

    # APP Init
    FileTwin(
        file_name=join(to_snake_case(app_name), "__init__.py"),
        content=f"from . import {types_name}",
    ).save(py_path)

    # APP API
    api_cls_name = to_camel_case("I " + app_name)
    methods = _get_oas_methods(oas)
    api_module_name = to_snake_case(f"{app_name}.i_{app_name}")
    api = FileTwin(
        file_name=module_to_path(api_module_name),
        content=_generate_api(methods, api_cls_name, types_module_name),
    ).save(py_path)

    # APP API Impl
    impl_cls_name = to_camel_case(app_name + " impl")
    impl_module_name = to_snake_case(f"{app_name}.{app_name}_impl")
    cls_str = _generate_api_impl(
        methods, api_module_name, types_module_name, api_cls_name, impl_cls_name
    )
    api_impl = FileTwin(
        file_name=module_to_path(impl_module_name), content=cls_str
    ).save(py_path)

    return RuntimeDomain(
        app_name=app_name,
        toolguard_common=common,
        app_types=types,
        app_api_class_name=api_cls_name,
        app_api=api,
        app_api_impl_class_name=impl_cls_name,
        app_api_impl=api_impl,
        app_api_size=len(methods),
    )


def _get_oas_methods(oas: OpenAPI):
    methods = []
    for path, path_item in oas.paths.items():  # noqa: B007
        path_item = oas.resolve_ref(path_item, PathItem)
        assert path_item
        for mtd, op in path_item.operations.items():  # noqa: B007
            op = oas.resolve_ref(op, Operation)
            if not op:
                continue
            params = (path_item.parameters or []) + (op.parameters or [])
            params = [oas.resolve_ref(p, Parameter) for p in params]
            args, ret = _make_signature(op, params, oas)  # type: ignore
            args_str = ", ".join(["self"] + [f"{arg}:{type}" for arg, type in args])
            sig = f"({args_str})->{ret}"

            body = f"return self._delegate.invoke('{to_snake_case(op.operationId)}', {ARGS}.model_dump(), {ret})"
            # if orign_funcs:
            #     func = find(orign_funcs or [], lambda fn: fn.__name__ == op.operationId) # type: ignore
            #     if func:
            #         body = _call_fn_body(func)
            methods.append(
                {
                    "name": to_snake_case(op.operationId),  # type: ignore
                    "signature": sig,
                    "doc": op.description,
                    "body": body,
                }
            )
    return methods


# def _call_fn_body(func:Callable):
#     module = inspect.getmodule(func)
#     if module is None or not hasattr(module, '__file__'):
#         raise ValueError("Function must be from an importable module")

#     module_name = module.__name__
#     qualname = func.__qualname__
#     func_name = func.__name__
#     parts = qualname.split('.')

#     if len(parts) == 1: # Regular function
#         return f"""
#     mod = importlib.import_module("{module_name}")
#     func = getattr(mod, "{func_name}")
#     return func(locals())"""

#     if len(parts) == 2:  # Classmethod or staticmethod
#         class_name = parts[0]
#         return f"""
#     mod = importlib.import_module("{module_name}")
#     cls = getattr(mod, "{class_name}")
#     func = getattr(cls, "{func_name}")
#     return func(locals())"""

#     if len(parts) > 2: # Instance method
#         class_name = parts[-2]
#         return f"""
#     mod = importlib.import_module("{module_name}")
#     cls = getattr(mod, "{class_name}")
#     instance = cls()
#     func = getattr(instance, "{func_name}")
#     return func(locals())"""
#     raise NotImplementedError("Unsupported function type or nested depth")


def _generate_api(methods: List, cls_name: str, types_module: str) -> str:
    return load_template("api.j2").render(
        types_module=types_module, class_name=cls_name, methods=methods
    )


def _generate_api_impl(
    methods: List, api_module: str, types_module: str, api_cls_name: str, cls_name: str
) -> str:
    return load_template("api_impl.j2").render(
        api_cls_name=api_cls_name,
        types_module=types_module,
        api_module=api_module,
        class_name=cls_name,
        methods=methods,
    )


def _make_signature(
    op: Operation, params: List[Parameter], oas: OpenAPI
) -> Tuple[Tuple[str, str], str]:
    fn_name = to_camel_case(op.operationId)
    args = []

    for param in params:
        if param.in_ == ParameterIn.path:
            args.append((param.name, _oas_to_py_type(param.schema_, oas) or "Any"))

    if find(params, lambda p: p.in_ == ParameterIn.query):
        query_type = f"{fn_name}ParametersQuery"
        args.append((ARGS, query_type))

    req_body = oas.resolve_ref(op.requestBody, RequestBody)
    if req_body:
        scm_or_ref = req_body.content_json.schema_
        body_type = _oas_to_py_type(scm_or_ref, oas)
        if body_type is None:
            body_type = f"{fn_name}Request"
        args.append((ARGS, body_type))

    rsp_or_ref = op.responses.get("200")
    rsp = oas.resolve_ref(rsp_or_ref, Response)
    if rsp:
        scm_or_ref = rsp.content_json.schema_
        if scm_or_ref:
            rsp_type = _oas_to_py_type(scm_or_ref, oas)
            if rsp_type is None:
                rsp_type = f"{fn_name}Response"
        else:
            rsp_type = "Any"
    return args, rsp_type


def _oas_to_py_type(scm_or_ref: Union[Reference, JSchema], oas: OpenAPI) -> str | None:
    if isinstance(scm_or_ref, Reference):
        typ = scm_or_ref.ref.split("/")[-1]
        return to_pascal_case(typ)

    scm = oas.resolve_ref(scm_or_ref, JSchema)
    if scm:
        py_type = _primitive_jschema_types_to_py(scm.type, scm.format)
        if py_type:
            return py_type
        # if scm.type == JSONSchemaTypes.array and scm.items:
        #     return f"List[{oas_to_py_type(scm.items, oas) or 'Any'}]"


def _primitive_jschema_types_to_py(
    type: Optional[str], format: Optional[str]
) -> Optional[str]:
    # https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.2.md#data-types
    if type == "string":
        if format == "date":
            return "datetime.date"
        if format == "date-time":
            return "datetime.datetime"
        if format in ["byte", "binary"]:
            return "bytes"
        return "str"
    if type == "integer":
        return "int"
    if type == "number":
        return "float"
    if type == "boolean":
        return "bool"
    return None



import os
from os.path import join
from typing import Callable, List, Optional, cast
import json
import logging

from langchain_core.tools import BaseTool


from .llm.i_tg_llm import I_TG_LLM
from .runtime import ToolGuardsCodeGenerationResult
from .data_types import ToolGuardSpec, load_tool_policy, ToolInfo
from .gen_py.gen_toolguards import generate_toolguards_from_functions, generate_toolguards_from_openapi
from .gen_spec.oas_summary import OASSummarizer
from .gen_spec.spec_generator import extract_toolguard_specs
from .common.langchain import langchain_tools_to_openapi

logger = logging.getLogger(__name__)

#Step1 only
async def generate_guard_specs(
	policy_text: str,
    tools: List[Callable] | List[BaseTool] | str,
	llm: I_TG_LLM,
    work_dir: str,
	short = False
)->List[ToolGuardSpec]:
	os.makedirs(work_dir, exist_ok=True)
	
	# case1: path to OpenAPI spec
	oas_path = tools if isinstance(tools, str) else None

	# case2: List of Langchain tools
	if isinstance(tools, list) and all([isinstance(tool, BaseTool) for tool in tools]):
		oas = langchain_tools_to_openapi(tools) # type: ignore
		oas_path = join(work_dir, "oas.json")
		oas.save(oas_path)

	if oas_path: # for cases 1 and 2
		with open(oas_path, 'r', encoding='utf-8') as file:
			oas = json.load(file)
		summarizer = OASSummarizer(oas)
		tools_info = summarizer.summarize()
		return await extract_toolguard_specs(policy_text, tools_info, work_dir, llm, None, short)
	
	# Case 3: List of functions + case 4: List of methods
	if isinstance(tools, list):
		tools_info = [ToolInfo.from_function(tool) for tool in tools]
		return await extract_toolguard_specs(policy_text, tools_info, work_dir, llm, None, short)
	
	raise NotImplementedError()


#Step2 only
async def generate_guards_from_specs(
		tools: List[Callable] | List[BaseTool] | str,
		tool_specs: List[ToolGuardSpec],
		work_dir: str,
		llm: I_TG_LLM,
		app_name: str = "myapp",
		lib_names: Optional[List[str]] = None,
		tool_names: Optional[List[str]] = None) -> ToolGuardsCodeGenerationResult:
	
	tool_specs = [policy for policy in tool_specs if (not tool_names) or (policy.tool_name in tool_names)]
	os.makedirs(work_dir, exist_ok=True)

	# case1: path to OpenAPI spec
	oas_path = tools if isinstance(tools, str) else None

	# case2: List of Langchain tools
	if isinstance(tools, list) and all([isinstance(tool, BaseTool) for tool in tools]):
		oas = langchain_tools_to_openapi(tools) # type: ignore
		oas_path = join(work_dir, "oas.json")
		oas.save(oas_path)

	if oas_path: # for cases 1 and 2
		return await generate_toolguards_from_openapi(app_name, tool_specs, work_dir, oas_path, llm)

	# Case 3: List of functions + case 4: List of methods
	funcs = [cast(Callable, tool) for tool in tools]
	return await generate_toolguards_from_functions(app_name, tool_specs, work_dir, funcs=funcs, llm=llm, module_roots=lib_names)
	

#load step1 results
def load_policies_in_folder(folder:str)->List[ToolGuardSpec]:
	files = [f for f in os.listdir(folder) 
		if os.path.isfile(join(folder, f)) and f.endswith(".json")]
	tool_policies = []
	for file in files:
		tool_name = file[:-len(".json")]
		policy = load_tool_policy(join(folder, file), tool_name)
		if policy.policy_items:
			tool_policies.append(policy)
	return tool_policies


# #Step1 and immediatly step2
# async def build_toolguards(
# 		policy_text:str, 
# 		tools: List[Callable]|str, 
# 		step1_out_dir:str, 
# 		step2_out_dir:str, 
# 		step1_llm:I_TG_LLM,
# 		step2_llm: I_TG_LLM, 
# 		app_name:str= "my_app", 
# 		tools2run: List[str] | None=None, 
# 		short1=True)->ToolGuardsCodeGenerationResult:
	
# 	os.makedirs(step1_out_dir, exist_ok=True)
# 	os.makedirs(step2_out_dir, exist_ok=True)

# 	# case1: path to OpenAPI spec
# 	oas_path = tools if isinstance(tools, str) else None

# 	# case2: List of Langchain tools
# 	if isinstance(tools, list) and all([isinstance(tool, BaseTool) for tool in tools]):
# 		oas = langchain_tools_to_openapi(tools) # type: ignore
# 		oas_path = f"{step1_out_dir}/oas.json"
# 		oas.save(oas_path)

# 	if oas_path: # for cases 1 and 2
# 		with open(oas_path, 'r', encoding='utf-8') as file:
# 			oas = json.load(file)
# 		summarizer = OASSummarizer(oas)
# 		tools_info = summarizer.summarize()
# 		tool_policies = await extract_toolguard_specs(policy_text, tools_info, step1_out_dir, step1_llm, tools2run, short1)
# 		return await generate_guards_from_specs(oas_path, tool_policies, step2_out_dir, llm=step2_llm, app_name=app_name, lib_names=None, tool_names=tools2run)
	
# 	# Case 3: List of functions + case 4: List of methods
# 	if isinstance(tools, list) and not oas_path:
# 		tools_info = [ToolInfo.from_function(tool) for tool in tools]
# 		tool_policies = await extract_toolguard_specs(policy_text, tools_info, step1_out_dir, step1_llm, tools2run, short1)
# 		return await generate_guards_from_specs(tools, tool_policies, step2_out_dir, llm=step2_llm, app_name=app_name, lib_names=None, tool_names=tools2run)
	
# 	raise Exception("Unknown tools")




from typing import List
from ...data_types import Domain, FileTwin, ToolGuardSpecItem
from mellea import generative


@generative
def generate_init_tests(
    fn_src: FileTwin,
    policy_item: ToolGuardSpecItem,
    domain: Domain,
    dependent_tool_names: List[str],
) -> str:
    """
        Generate Python unit tests for a function to verify tool-call compliance with policy constraints.

        Args:
            fn_src (FileTwin): Source code containing the function-under-test signature.
            policy_item (ToolGuardSpecItem): Specification of the function-under-test, including positive and negative examples.
            domain (Domain): available data types and interfaces the test can use.
            dependent_tool_names (List[str]): other tool names that this tool depends on.

        Returns:
            str: Generated Python unit test code.

        This function creates unit tests to validate the behavior of a given function-under-test.
        The function-under-test checks the argument data, and raise an exception if they violated the requirements in the policy item.

        Test Generation Rules:
        - Make sure to Python import all items in fn_src, common and domain modules.
        - A `policy_item` has multiple `compliance_examples` and `violation_examples` examples.
            - For each `compliance_examples`, ONE test method is generated.
            - For each `violation_examples`, ONE test method is generated.
                - The function-under-test is EXPECTED to raise a `PolicyViolationException`.
                - use `with pytest.raises(PolicyViolationException): function_under_test()` to expect for exceptions.

            - Test class and method names should be meaningful and use up to **six words in snake_case**.
            - For each test, add a comment quoting the policy item case that this function is testing
            - Failure message should describe the test scenario that failed, the expected and the actual outcomes.

        Data population and references:
        - For compliance examples, populate all fields.
            - For collections (arrays, dict and sets) populate at least one item.
        - You should `MagicMock` to mock the return_value from ALL tools listed in `dependent_tool_names`.
            - Use `MagicMock.side_effect` function to return the expected value only when the expected parameters are passed.
            - Do not use `MagicMock.return_value`.
        - For time dependent attributes, compute the timestamp dynamically (avoid hardcoded times).
            - for example, to set a timestamp occurred 24 hours ago, use something like: `created_at = (datetime.now() - timedelta(hours=24)).strftime("%Y-%m-%dT%H:%M:%S")`.
            - import the required date and time libraries. for example: `from datetime import datetime, timedelta`
        - If you have a choice passing a plain a Pydantic model or a `Dictionary`, prefer Pydantic.

        Example:
        * fn_src:
    ```python
    # file: my_app/guard_create_reservation.py
    def guard_create_reservation(api: SomeAPI, user_id: str, hotel_id: str, reservation_date: str, persons: int):
        ...
    ```
        * policy_item.description = "cannot book a room for a date in the past"
        * policy_item.violation_examples = ["book a room for a hotel, one week ago"]
        * Dependent_tool_names: `["get_user", "get_hotel"]`
        * Domain:
    ```python
    # file: my_app/api.py
    class SomeAPI(ABC):
        def get_user(self, user_id):
            ...
        def get_hotel(self, hotel_id):
            ...
        def create_reservation(self, user_id: str, hotel_id: str, reservation_date: str, persons: int):
            \"\"\"
            Args:
                ...
                reservation_date: check in date, in `YYYY-MM-DDTHH:MM:SS` format
            \"\"\"
            ...
    ```

        Should return this snippet:
    ```python
    from unittest.mock import MagicMock
    import pytest
    from toolguard.data_types import PolicyViolationException
    from my_app.guard_create_reservation import guard_create_reservation
    from my_app.api import *

    def test_violation_book_room_in_the_past():
        \"\"\"
        Policy: "cannot book room for a date in the past"
        Example: "book a room for a hotel, one week ago"
        \"\"\"

        # mock other tools function return values
        user = User(user_id="123", ...)
        hotel = Hotel(hotel_id="789", ...)

        api = MagicMock(spec=SomeAPI)
        api.get_user.side_effect = lambda user_id: user if user_id == "123" else None
        api.get_hotel.side_effect = lambda hotel_id: hotel if hotel_id == "789" else None

        #invoke function under test.
        with pytest.raises(PolicyViolationException):
            next_week = (datetime.now() + timedelta(days=7)).strftime("%Y-%m-%dT%H:%M:%S")
            guard_create_reservation(api, user_id="123", hotel_id="789", reservation_date=next_week, persons=3)
    ```
    """
    ...


@generative
def improve_tests(
    prev_impl: str,
    domain: Domain,
    policy_item: ToolGuardSpecItem,
    review_comments: List[str],
    dependent_tool_names: List[str],
) -> str:
    """
    Improve the previous test functions (in Python) to check the given tool policy-items according to the review-comments.

    Args:
        prev_impl (str): previous implementation of a Python function.
        domain (Domain): Python source code defining available data types and APIs that the test can use.
        tool (ToolGuardSpecItem): Requirements for this tool.
        review_comments (List[str]): Review comments on the current implementation. For example, pylint errors or Failed unit-tests.
        dependent_tool_names (List[str]): other tool names that this tool depends on.

    Returns:
        str: Improved implementation pytest test functions.

    Implementation Rules:
    - Do not change the function signatures.
    - You can add import statements, but dont remove them.
    """
    ...



import json
from typing import Any, Dict, List

from ..data_types import ToolInfo


class OASSummarizer:
    def __init__(self, oas: Dict[str, Any]):
        self.oas = oas
        self.components = oas.get("components", {}).get("schemas", {})

    def summarize(self) -> List[ToolInfo]:
        operations = []
        for path, methods in self.oas.get("paths", {}).items():
            for method, operation in methods.items():
                summary = self._format_operation(path, method, operation)
                tool_info = ToolInfo(
                    name=summary.get("name"),
                    description=summary.get("description"),
                    signature=summary.get("signature"),
                    parameters=summary.get("params"),
                    full_description=json.dumps(summary, indent=4),
                )
                operations.append(tool_info)
        return operations

    def _format_operation(
        self, path: str, method: str, operation: Dict[str, Any]
    ) -> Dict[str, Any]:
        operation_id = operation.get(
            "operationId", f"{method}_{path.strip('/').replace('/', '_')}"
        )
        class_name = operation_id
        description = operation.get("description", "")
        request_body = operation.get("requestBody", {})
        params = self._parse_request_body(request_body) if request_body else {}
        signature = self._generate_signature(
            class_name, params, operation.get("responses", {})
        )
        example = operation.get(
            "x-input-examples", []
        )  # self._generate_example(class_name, params)
        output_examples = operation.get(
            "x-output-examples", []
        )  # self._parse_response_examples(operation.get("responses", {}))

        return {
            "name": class_name,
            "signature": signature,
            "description": description,
            "params": params,
            "examples": [example],
            "output_examples": output_examples,
        }

    def _parse_request_body(self, request_body: Dict[str, Any]) -> Dict[str, Any]:
        content = request_body.get("content", {}).get("application/json", {})
        schema = self._resolve_ref(content.get("schema", {}))
        props = schema.get("properties", {})
        required = schema.get("required", [])
        params = {}
        for param_name, param_schema in props.items():
            resolved_schema = self._resolve_ref(param_schema)
            param_type = self._resolve_schema_type(resolved_schema)
            param_desc = resolved_schema.get("description", "")
            params[param_name] = {
                "type": param_type,
                "description": param_desc,
                "required": param_name in required,
            }
        return params

    # def _resolve_schema_type(self, schema: Dict[str, Any]) -> str:
    #     if "anyOf" in schema:
    #         return "Union[" + ", ".join(self._resolve_schema_type(s) for s in schema["anyOf"]) + "]"
    #     if "oneOf" in schema:
    #         return "Union[" + ", ".join(self._resolve_schema_type(s) for s in schema["oneOf"]) + "]"
    #     if "$ref" in schema:
    #         return self._resolve_schema_type(self._resolve_ref(schema))
    #     if schema.get("type") == "array":
    #         item_type = self._resolve_schema_type(schema.get("items", {}))
    #         return f"List[{item_type}]"
    #     if schema.get("type") == "object":
    #         return "Dict[str, Any]"
    #
    #     return {
    #         "string": "str",
    #         "integer": "int",
    #         "number": "float",
    #         "boolean": "bool",
    #         "object": "Dict[str, Any]",
    #     }.get(schema.get("type", "Any"), "Any")

    def _resolve_schema_type(self, schema: Dict[str, Any]) -> str:
        if "anyOf" in schema:
            return (
                "Union["
                + ", ".join(self._resolve_schema_type(s) for s in schema["anyOf"])
                + "]"
            )
        if "oneOf" in schema:
            return (
                "Union["
                + ", ".join(self._resolve_schema_type(s) for s in schema["oneOf"])
                + "]"
            )
        if "$ref" in schema:
            return self._resolve_schema_type(self._resolve_ref(schema))
        if schema.get("type") == "array":
            item_type = self._resolve_schema_type(schema.get("items", {}))
            return f"List[{item_type}]"
        if schema.get("type") == "object":
            return "Dict[str, Any]"

        type_value = schema.get("type", "Any")
        if isinstance(type_value, list):
            # Filter out "null" and resolve remaining types
            non_null_types = [t for t in type_value if t != "null"]
            if not non_null_types:
                return "Optional[Any]"
            if len(non_null_types) == 1:
                base_type = self._resolve_schema_type(
                    {**schema, "type": non_null_types[0]}
                )
            else:
                base_type = (
                    "Union["
                    + ", ".join(
                        self._resolve_schema_type({"type": t}) for t in non_null_types
                    )
                    + "]"
                )
            return f"Optional[{base_type}]" if "null" in type_value else base_type

        return {
            "string": "str",
            "integer": "int",
            "number": "float",
            "boolean": "bool",
            "object": "Dict[str, Any]",
        }.get(type_value, "Any")

    def _resolve_ref(self, schema: Dict[str, Any]) -> Dict[str, Any]:
        if isinstance(schema, Dict):
            if "$ref" in schema:
                ref_path = schema["$ref"]
                if ref_path.startswith("#/components/schemas/"):
                    key = ref_path.split("/")[-1]
                    return self.components.get(key, {})
            return schema
        return schema

    def _generate_signature(
        self, class_name: str, params: Dict[str, Any], responses: Dict[str, Any]
    ) -> str:
        args = ", ".join(f"{name}: {meta['type']}" for name, meta in params.items())
        out = "str"
        if responses and "200" in responses:
            content = responses["200"]["content"]
            app_json = content.get("application/json", {})
            schema = self._resolve_ref(app_json.get("schema", {}))
            out = self._resolve_schema_type(schema)
        return f"{class_name}({args}) -> {out}"

    def _generate_example(self, class_name: str, params: Dict[str, Any]) -> str:
        args = ", ".join(
            '"example_string"' if meta["type"].startswith("str") else "0"
            for _, meta in params.items()
        )
        return f"{class_name}({args})"

    def _parse_response_examples(self, responses: Dict[str, Any]) -> List[str]:
        examples = []
        for response in responses.values():
            content = response.get("content", {})
            app_json = content.get("application/json", {})
            schema = self._resolve_ref(app_json.get("schema", {}))

            if "example" in app_json:
                example_data = app_json["example"]
            elif "examples" in app_json:
                example_data = next(iter(app_json["examples"].values()), {}).get(
                    "value"
                )
            else:
                example_data = self._construct_example_from_schema(schema)

            if example_data is not None:
                try:
                    examples.append(json.dumps(example_data))
                except Exception:
                    examples.append(str(example_data))
        return examples

    def _construct_example_from_schema(self, schema: Dict[str, Any]) -> Any:
        schema = self._resolve_ref(schema)
        if not isinstance(schema, Dict):
            return schema
        schema_type = schema.get("type")

        if schema_type == "object":
            if "additionalProperties" in schema:
                value_schema = self._resolve_ref(schema["additionalProperties"])
                return {
                    "example_key": self._construct_example_from_schema(value_schema)
                }
            props = schema.get("properties", {})
            return {
                key: self._construct_example_from_schema(self._resolve_ref(subschema))
                for key, subschema in props.items()
            }

        if schema_type == "array":
            item_schema = self._resolve_ref(schema.get("items", {}))
            return [self._construct_example_from_schema(item_schema)]

        if schema_type == "string":
            return schema.get("example", "example_string")
        if schema_type == "integer":
            return schema.get("example", 42)
        if schema_type == "number":
            return schema.get("example", 3.14)
        if schema_type == "boolean":
            return schema.get("example", True)

        if "anyOf" in schema:
            return self._construct_example_from_schema(schema["anyOf"][0])
        if "oneOf" in schema:
            return self._construct_example_from_schema(schema["oneOf"][0])
        if "allOf" in schema:
            return self._construct_example_from_schema(schema["allOf"][0])

        return "example_value"


# if __name__ == "__main__":
#     oas_file = (
#         "/Users/naamazwerdling/Documents/OASB/policy_validation/airline/oas2.json"
#     )
#     # oas_file = "/Users/naamazwerdling/Documents/OASB/policy_validation/orca/bank/oas.json"
#     shortfile = "/Users/naamazwerdling/Documents/OASB/policy_validation/airline/s1.json"
#     # shortfile = "/Users/naamazwerdling/Documents/OASB/policy_validation/orca/bank/short.json"
#     with open(oas_file) as f:
#         oas_data = json.load(f)

#     summarizer = OASSummarizer(oas_data)
#     summary = summarizer.summarize()
#     with open(shortfile, "w") as outfile:
#         json.dump(summary, outfile, indent=4)

#     print(json.dumps(summary, indent=2))


from importlib import import_module
import inspect
import os
import asyncio
import logging
from os.path import join
import re
from typing import Callable, List, Tuple

from mellea import MelleaSession

from ..common import py
from ..common.llm_py import get_code_content
from ..common.py_doc_str import extract_docstr_args
from ..common.str import to_snake_case
from ..data_types import (
    DEBUG_DIR,
    TESTS_DIR,
    FileTwin,
    RuntimeDomain,
    ToolGuardSpec,
    ToolGuardSpecItem,
)
from ..gen_py.consts import (
    guard_fn_module_name,
    guard_fn_name,
    guard_item_fn_module_name,
    guard_item_fn_name,
    test_fn_module_name,
)
from ..runtime import ToolGuardCodeResult, find_class_in_module
from ..gen_py.tool_dependencies import tool_dependencies
from ..gen_py.prompts.gen_tests import generate_init_tests, improve_tests
from ..gen_py.prompts.improve_guard import improve_tool_guard
from ..gen_py.templates import load_template
from .utils import pytest
from .utils import pyright

logger = logging.getLogger(__name__)

MAX_TOOL_IMPROVEMENTS = 5
MAX_TEST_GEN_TRIALS = 3

class ToolGuardGenerator:
    app_name: str
    py_path: str
    tool_policy: ToolGuardSpec
    domain: RuntimeDomain
    common: FileTwin

    def __init__(
        self,
        app_name: str,
        tool_policy: ToolGuardSpec,
        py_path: str,
        domain: RuntimeDomain,
        m: MelleaSession
    ) -> None:
        self.py_path = py_path
        self.app_name = app_name
        self.tool_policy = tool_policy
        self.domain = domain
        self.m = m

    def start(self):
        app_path = join(self.py_path, to_snake_case(self.app_name))
        os.makedirs(app_path, exist_ok=True)
        os.makedirs(
            join(app_path, to_snake_case(self.tool_policy.tool_name)), exist_ok=True
        )
        os.makedirs(join(self.py_path, to_snake_case(DEBUG_DIR)), exist_ok=True)
        os.makedirs(
            join(
                self.py_path,
                to_snake_case(DEBUG_DIR),
                to_snake_case(self.tool_policy.tool_name),
            ),
            exist_ok=True,
        )
        for item in self.tool_policy.policy_items:
            os.makedirs(
                join(
                    self.py_path,
                    to_snake_case(DEBUG_DIR),
                    to_snake_case(self.tool_policy.tool_name),
                    to_snake_case(item.name),
                ),
                exist_ok=True,
            )
        os.makedirs(join(self.py_path, to_snake_case(TESTS_DIR)), exist_ok=True)

    async def generate(self) -> ToolGuardCodeResult:
        self.start()
        
        with py.temp_python_path(self.py_path):
            tool_guard, init_item_guards = self._create_initial_tool_guards()
    
            # Generate guards for all tool items
            tests_and_guards = await asyncio.gather(* [
                self._generate_item_tests_and_guard(item, item_guard)
                    for item, item_guard in zip(self.tool_policy.policy_items, init_item_guards)
            ])

        item_tests, item_guards = zip(*tests_and_guards)
        return ToolGuardCodeResult(
            tool=self.tool_policy,
            guard_fn_name=guard_fn_name(self.tool_policy),
            guard_file=tool_guard,
            item_guard_files=item_guards,
            test_files=item_tests,
        )

    async def _generate_item_tests_and_guard(
        self, item: ToolGuardSpecItem, init_guard: FileTwin
    ) -> Tuple[FileTwin | None, FileTwin]:
        # Dependencies of this tool
        tool_fn_name = to_snake_case(self.tool_policy.tool_name)
        tool_fn = self._find_api_function(tool_fn_name)
        sig_str = f"{tool_fn_name}{str(inspect.signature(tool_fn))}"
        dep_tools = []
        if self.domain.app_api_size > 1:
            domain = self.domain.get_definitions_only()  # remove runtime fields
            dep_tools = list(await tool_dependencies(item, sig_str, domain, self.m))
        logger.debug(f"Dependencies of '{item.name}': {dep_tools}")

        # Generate tests
        try:
            guard_tests = await self._generate_tests(item, init_guard, dep_tools)
        except Exception as ex:
            logger.warning(f"Tests generation failed for item {item.name} %s", str(ex))
            try:
                logger.warning("try to generate the code without tests... %s", str(ex))
                guard = await self._improve_guard(item, init_guard, [], dep_tools)
                return None, guard
            except Exception as ex:
                logger.exception(ex)
                logger.warning(
                    "guard generation failed. returning initial guard: %s", str(ex)
                )
                return None, init_guard

        # Tests generated, now generate guards
        try:
            guard = await self._improve_guard_green_loop(
                item, init_guard, guard_tests, dep_tools
            )
            logger.debug(
                f"tool item generated successfully '{item.name}'"
            )  # 😄🎉 Happy path
            return guard_tests, guard
        except Exception as ex:
            logger.exception(ex)
            logger.warning(
                "guard generation failed. returning initial guard: %s", str(ex)
            )
            return None, init_guard

    # async def tool_dependencies(self, policy_item: ToolGuardSpecItem, tool_signature: str) -> Set[str]:
    #     domain = self.domain.get_definitions_only() #remove runtime fields
    #     pseudo_code = await tool_policy_pseudo_code(policy_item, tool_signature, domain)
    #     dep_tools = await extract_api_dependencies_from_pseudo_code(pseudo_code, domain)
    #     return set(dep_tools)

    async def _generate_tests(
        self, item: ToolGuardSpecItem, guard: FileTwin, dep_tools: List[str]
    ) -> FileTwin:
        test_file_name = join(
            TESTS_DIR, self.tool_policy.tool_name, f"{test_fn_module_name(item)}.py"
        )
        errors = []
        test_file = None
        trials = "a b c".split()
        for trial_no in trials:
            logger.debug(
                f"Generating tests iteration '{trial_no}' for tool {self.tool_policy.tool_name} '{item.name}'."
            )
            domain = self.domain.get_definitions_only()  # remove runtime fields
            first_time = trial_no == "a"
            if first_time:
                res = generate_init_tests(
                    self.m,
                    fn_src=guard,
                    policy_item=item,
                    domain=domain,  # noqa: B023
                    dependent_tool_names=dep_tools,
                )
            else:
                assert test_file
                res = improve_tests(
                    self.m,
                    prev_impl=test_file.content,  # noqa: B023
                    domain=domain,  # noqa: B023
                    policy_item=item,
                    review_comments=errors,  # noqa: B023
                    dependent_tool_names=dep_tools,
                )

            test_file = FileTwin(
                file_name=test_file_name, content=get_code_content(res)
            ).save(self.py_path)
            test_file.save_as(self.py_path, self.debug_dir(item, f"test_{trial_no}.py"))

            syntax_report = pyright.run(self.py_path, test_file.file_name)
            FileTwin(
                    file_name= self.debug_dir(item, f"test_{trial_no}_pyright.json"),
                    content=syntax_report.model_dump_json(indent=2)
                ).save(self.py_path)
            
            if syntax_report.summary.errorCount>0:
                logger.warning(f"{syntax_report.summary.errorCount} syntax errors in tests iteration '{trial_no}' in item '{item.name}'.")
                errors = syntax_report.list_error_messages(test_file.content)
                continue

            # syntax ok, try to run it...
            logger.debug(
                f"Generated Tests for tool '{self.tool_policy.tool_name}' '{item.name}'(trial='{trial_no}')"
            )
            return test_file
            # report_file_name = self.debug_dir(item, f"test_{trial_no}_pytest.json")
            # pytest_report = pytest.run(
            #     self.py_path, test_file.file_name, report_file_name
            # )
            # if (
            #     pytest_report.all_tests_collected_successfully()
            #     and pytest_report.non_empty_tests()
            # ):
            #     return test_file
            # if not pytest_report.non_empty_tests():  # empty test set
            #     errors = ["empty set of generated unit tests is not allowed"]
            # else:
            #     errors = pytest_report.list_errors()

        raise Exception("Generated tests contain syntax errors")

    async def _improve_guard_green_loop(
        self,
        item: ToolGuardSpecItem,
        guard: FileTwin,
        tests: FileTwin,
        dep_tools: List[str],
    ) -> FileTwin:
        trial_no = 0
        while trial_no < MAX_TOOL_IMPROVEMENTS:
            pytest_report_file = self.debug_dir(item, f"guard_{trial_no}_pytest.json")
            errors = pytest.run(
                self.py_path, tests.file_name, pytest_report_file
            ).list_errors()
            if errors:
                logger.debug(f"'{item.name}' guard function tests failed. Retrying...")

                trial_no += 1
                try:
                    guard = await self._improve_guard(
                        item, guard, errors, dep_tools, trial_no
                    )
                except Exception:
                    continue  # probably a syntax error in the generated code. lets retry...
            else:
                logger.debug(
                    f"'{item.name}' guard function generated succefully and is Green 😄🎉. "
                )
                return guard  # Green

        raise Exception(
            f"Failed {MAX_TOOL_IMPROVEMENTS} times to generate guard function for tool {to_snake_case(self.tool_policy.tool_name)} policy: {item.name}"
        )

    async def _improve_guard(
        self,
        item: ToolGuardSpecItem,
        prev_guard: FileTwin,
        review_comments: List[str],
        dep_tools: List[str],
        round: int = 0,
    ) -> FileTwin:
        module_name = guard_item_fn_module_name(item)
        errors = []
        trials = "a b c".split()
        for trial in trials:
            logger.debug(
                f"Improving guard function '{module_name}'... (trial = {round}.{trial})"
            )
            domain = self.domain.get_definitions_only()  # omit runtime fields
            prev_python = get_code_content(prev_guard.content)
            res = improve_tool_guard(
                self.m,
                prev_impl=prev_python,  # noqa: B023
                domain=domain,  # noqa: B023
                policy_item=item,
                dependent_tool_names=dep_tools,
                review_comments=review_comments + errors,  # noqa: B023
            )

            guard = FileTwin(
                file_name=prev_guard.file_name, content=get_code_content(res)
            ).save(self.py_path)
            guard.save_as(
                self.py_path, self.debug_dir(item, f"guard_{round}_{trial}.py")
            )

            syntax_report = pyright.run(self.py_path, guard.file_name)
            FileTwin(
                file_name=self.debug_dir(item, f"guard_{round}_{trial}.pyright.json"),
                content=syntax_report.model_dump_json(indent=2),
            ).save(self.py_path)
            logger.info(
                f"Generated function {module_name} with {syntax_report.summary.errorCount} errors."
            )

            if syntax_report.summary.errorCount > 0:
                # Syntax errors. retry...
                errors = syntax_report.list_error_messages(guard.content)
                continue

            guard.save_as(self.py_path, self.debug_dir(item, f"guard_{round}_final.py"))
            return (
                guard  # Happy path. improved vesion of the guard with no syntax errors
            )

        # Failed to generate valid python after iterations
        raise Exception(f"Syntax error generating for tool '{item.name}'.")

    def _find_api_function(self, tool_fn_name: str):
        module = import_module(py.path_to_module(self.domain.app_api.file_name))
        assert module, f"File not found {self.domain.app_api.file_name}"
        cls = find_class_in_module(module, self.domain.app_api_class_name)
        return getattr(cls, tool_fn_name)

    def _create_initial_tool_guards(self) -> Tuple[FileTwin, List[FileTwin]]:
        tool_fn_name = to_snake_case(self.tool_policy.tool_name)
        tool_fn = self._find_api_function(tool_fn_name)
        assert tool_fn, f"Function not found, {tool_fn_name}"

        # __init__.py
        path = join(to_snake_case(self.app_name), tool_fn_name, "__init__.py")
        FileTwin(file_name=path, content="").save(self.py_path)

        # item guards files
        item_files = [
            self._create_item_module(item, tool_fn)
            for item in self.tool_policy.policy_items
        ]
        # tool guard file
        tool_file = self._create_tool_module(tool_fn, item_files)

        # Save to debug folder
        for item_guard_fn, policy_item in zip(
            item_files, self.tool_policy.policy_items
        ):
            item_guard_fn.save_as(self.py_path, self.debug_dir(policy_item, "g0.py"))

        return (tool_file, item_files)

    def _create_tool_module(
        self, tool_fn: Callable, item_files: List[FileTwin]
    ) -> FileTwin:
        file_name = join(
            to_snake_case(self.app_name),
            to_snake_case(self.tool_policy.tool_name),
            py.py_extension(guard_fn_module_name(self.tool_policy)),
        )
        items = [
            {"guard_fn": guard_item_fn_name(item), "file_name": file.file_name}
            for (item, file) in zip(self.tool_policy.policy_items, item_files)
        ]
        sig = inspect.signature(tool_fn)
        sig_str = self._signature_str(sig)
        args_call = ", ".join([p for p in sig.parameters if p != "self"])
        args_doc_str = extract_docstr_args(tool_fn)
        extra_imports = []
        if "Decimal" in sig_str:
            extra_imports.append("from decimal import Decimal")

        return FileTwin(
            file_name=file_name,
            content=load_template("tool_guard.j2").render(
                domain=self.domain,
                method={
                    "name": guard_fn_name(self.tool_policy),
                    "signature": sig_str,
                    "args_call": args_call,
                    "args_doc_str": args_doc_str,
                },
                items=items,
                extra_imports=extra_imports,
            ),
        ).save(self.py_path)

    def _signature_str(self, sig: inspect.Signature):
        sig_str = str(sig)
        sig_str = sig_str[
            sig_str.find("self,") + len("self,") : sig_str.rfind(")")
        ].strip()
        # Strip module prefixes like airline.airline_types.XXX → XXX
        clean_sig_str = re.sub(r"\b(?:\w+\.)+(\w+)", r"\1", sig_str)
        return clean_sig_str

    def _create_item_module(
        self, tool_item: ToolGuardSpecItem, tool_fn: Callable
    ) -> FileTwin:
        file_name = join(
            to_snake_case(self.app_name),
            to_snake_case(self.tool_policy.tool_name),
            py.py_extension(guard_item_fn_module_name(tool_item)),
        )
        sig_str = self._signature_str(inspect.signature(tool_fn))
        args_doc_str = extract_docstr_args(tool_fn)
        extra_imports = []
        if "Decimal" in sig_str:
            extra_imports.append("from decimal import Decimal")
        return FileTwin(
            file_name=file_name,
            content=load_template("tool_item_guard.j2").render(
                domain=self.domain,
                method={
                    "name": guard_item_fn_name(tool_item),
                    "signature": sig_str,
                    "args_doc_str": args_doc_str,
                },
                policy=tool_item.description,
                extra_imports=extra_imports,
            ),
        ).save(self.py_path)

    def debug_dir(self, policy_item: ToolGuardSpecItem, dir: str):
        return join(
            DEBUG_DIR,
            to_snake_case(self.tool_policy.tool_name),
            to_snake_case(policy_item.name),
            dir,
        )



import importlib
import inspect
import json
import os
import sys
from types import ModuleType
from typing import Any, Dict, List, Optional, Type, Callable, TypeVar
from pydantic import BaseModel
from langchain_core.tools import BaseTool

import functools
from .data_types import API_PARAM, ARGS_PARAM, RESULTS_FILENAME, FileTwin, RuntimeDomain, ToolGuardSpec

from abc import ABC, abstractmethod


class IToolInvoker(ABC):
    T = TypeVar("T")

    @abstractmethod
    def invoke(
        self, toolname: str, arguments: Dict[str, Any], return_type: Type[T]
    ) -> T: ...

def load_toolguard_code_result(directory: str, filename: str = RESULTS_FILENAME):
    full_path = os.path.join(directory, filename)
    with open(full_path, "r", encoding="utf-8") as f:
        data = json.load(f)
    return ToolGuardsCodeGenerationResult(**data)

def load_toolguards(directory: str, filename: str = RESULTS_FILENAME) -> "ToolguardRuntime":
    return ToolguardRuntime(load_toolguard_code_result(directory, filename), directory)

class ToolGuardCodeResult(BaseModel):
    tool: ToolGuardSpec
    guard_fn_name: str
    guard_file: FileTwin
    item_guard_files: List[FileTwin | None]
    test_files: List[FileTwin | None]


class ToolGuardsCodeGenerationResult(BaseModel):
    out_dir: str
    domain: RuntimeDomain
    tools: Dict[str, ToolGuardCodeResult]

    def save(
        self, directory: str, filename: str = RESULTS_FILENAME
    ) -> "ToolGuardsCodeGenerationResult":
        full_path = os.path.join(directory, filename)
        with open(full_path, "w", encoding="utf-8") as f:
            json.dump(self.model_dump(), f, indent=2)
        return self


class ToolguardRuntime:
    _original_pypath:list[str] = []

    def __init__(self, result: ToolGuardsCodeGenerationResult, ctx_dir: str) -> None:
        self._ctx_dir = ctx_dir
        self._result = result

    def __enter__(self):
        #add folder to python path
        self._original_pypath = list(sys.path) #remember old path
        sys.path.insert(0, os.path.abspath(self._ctx_dir))

        #cache the tool guards
        self._guards = {}
        for tool_name, tool_result in self._result.tools.items():
            mod_name = file_to_module_name(tool_result.guard_file.file_name)
            module = importlib.import_module(mod_name)
            guard_fn = find_function_in_module(module, tool_result.guard_fn_name)
            assert guard_fn, "Guard not found"
            self._guards[tool_name] = guard_fn

        return self

    def __exit__(self, exc_type, exc, tb):
        del self._guards
        #back to original python path
        sys.path[:] = self._original_pypath
        return False
    
    def _make_args(self, guard_fn:Callable, args: dict,  delegate: IToolInvoker)->Dict[str, Any]:
        sig = inspect.signature(guard_fn)
        guard_args = {}
        for p_name, param in sig.parameters.items():
            if p_name == API_PARAM:
                mod_name = file_to_module_name(self._result.domain.app_api_impl.file_name)
                module = importlib.import_module(mod_name)
                clazz = find_class_in_module(module, self._result.domain.app_api_impl_class_name)
                assert clazz, f"class {self._result.domain.app_api_impl_class_name} not found in {self._result.domain.app_api_impl.file_name}"
                guard_args[p_name] = clazz(delegate)
            else:
                arg_val = args.get(p_name)
                if arg_val is None and p_name == ARGS_PARAM:
                    arg_val = args
                
                if inspect.isclass(param.annotation) and issubclass(param.annotation, BaseModel):
                    guard_args[p_name] = param.annotation.model_construct(**arg_val)
                else:
                    guard_args[p_name] = arg_val
        return guard_args

    def check_toolcall(self, tool_name: str, args: dict, delegate: IToolInvoker):
        guard_fn = self._guards.get(tool_name)
        if guard_fn is None:  # No guard assigned to this tool
            return
        guard_fn(**self._make_args(guard_fn, args, delegate))

def file_to_module_name(file_path:str):
    return file_path.removesuffix('.py').replace('/', '.')

def find_function_in_module(module: ModuleType, function_name: str):
    func = getattr(module, function_name, None)
    if func is None or not inspect.isfunction(func):
        raise AttributeError(
            f"Function '{function_name}' not found in module '{module.__name__}'"
        )
    return func


def find_class_in_module(module: ModuleType, class_name: str) -> Optional[Type]:
    cls = getattr(module, class_name, None)
    if isinstance(cls, type):
        return cls
    return None


T = TypeVar("T")


def guard_methods(obj: T, guards_folder: str) -> T:
    """Wraps all public bound methods of the given instance using the given wrapper."""
    for attr_name in dir(obj):
        if attr_name.startswith("_"):
            continue
        attr = getattr(obj, attr_name)
        if callable(attr):
            wrapped = guard_before_call(guards_folder)(attr)
            setattr(obj, attr_name, wrapped)
    return obj


class ToolMethodsInvoker(IToolInvoker):
    def __init__(self, object: object) -> None:
        self._obj = object

    def invoke(self, toolname: str, arguments: Dict[str, Any], return_type: Type[T]) -> T:
        mtd = getattr(self._obj, toolname)
        assert callable(mtd), f"Tool {toolname} was not found"
        return mtd(**arguments)


class ToolFunctionsInvoker(IToolInvoker):
    def __init__(self, funcs: List[Callable]) -> None:
        self._funcs_by_name = {func.__name__: func for func in funcs}

    def invoke(self, toolname: str, arguments: Dict[str, Any], return_type: Type[T]) -> T:
        func = self._funcs_by_name.get(toolname)
        assert callable(func), f"Tool {toolname} was not found"
        return func(**arguments)

class LangchainToolInvoker(IToolInvoker):
    T = TypeVar("T")
    _tools: Dict[str, BaseTool]

    def __init__(self, tools: List[BaseTool]) -> None:
        self._tools = {tool.name: tool for tool in tools}

    def invoke(self, toolname: str, arguments: Dict[str, Any], return_type: Type[T])->T:
        tool = self._tools.get(toolname)
        if tool:
            return tool.invoke(arguments)
        raise ValueError(f"unknown tool {toolname}")
    
def guard_before_call(guards_folder: str) -> Callable[[Callable], Callable]:
    """Decorator factory that logs function calls to the given logfile."""
    toolguards = load_toolguards(guards_folder)

    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            toolguards.check_toolcall(
                func.__name__, kwargs, ToolMethodsInvoker(func.__self__)
            )
            return func(*args, **kwargs)

        return wrapper  # type: ignore

    return decorator



import subprocess


def run(oas_file: str):
    # see https://github.com/koxudaxi/datamodel-code-generator
    res = subprocess.run(
        [
            "datamodel-codegen",
            "--use-field-description",
            "--use-schema-description",
            "--output-model-type",
            "pydantic_v2.BaseModel",  # "typing.TypedDict",
            "--collapse-root-models",
            # "--force-optional",
            "--reuse-model",
            "--enum-field-as-literal",
            "all",
            "--input-file-type",
            "openapi",
            "--use-operation-id-as-name",
            "--openapi-scopes",
            "paths",
            "parameters",
            "schemas",
            "--input",
            oas_file,
            # "--output", domain_file
        ],
        capture_output=True,
        text=True,
    )
    if res.returncode != 0:
        raise Exception(res.stderr)
    return res.stdout




from enum import StrEnum
import json
import os
from os.path import join
from typing import Any, List, Dict, Optional
from pydantic import BaseModel, Field

from ...common.multi_process import run_in_process
from ...common.safe_py import run_safe_python
from ...data_types import FileTwin




class TestOutcome(StrEnum):
    passed = "passed"
    failed = "failed"


class TracebackEntry(BaseModel):
    path: str
    lineno: int
    message: str


class CrashInfo(BaseModel):
    path: str
    lineno: int
    message: str


class CallInfo(BaseModel):
    duration: float
    outcome: TestOutcome
    crash: Optional[CrashInfo] = None
    traceback: Optional[List[TracebackEntry]] = None
    longrepr: Optional[str] = None


class TestPhase(BaseModel):
    duration: float
    outcome: TestOutcome


class TestResult(BaseModel):
    nodeid: str
    lineno: int
    outcome: TestOutcome
    keywords: List[str]
    setup: TestPhase
    call: CallInfo
    user_properties: Optional[List[Any]] = None
    teardown: TestPhase


class ResultEntry(BaseModel):
    nodeid: str
    type: str
    lineno: Optional[int] = None


class Collector(BaseModel):
    nodeid: str
    outcome: TestOutcome
    result: List[ResultEntry]
    longrepr: Optional[str] = None


class Summary(BaseModel):
    failed: Optional[int] = 0
    total: int
    collected: int


class TestReport(BaseModel):
    created: float
    duration: float
    exitcode: int
    root: str
    environment: Dict[str, str]
    summary: Summary
    collectors: List[Collector] = Field(default=[])
    tests: List[TestResult]

    def all_tests_passed(self) -> bool:
        return all([test.outcome == TestOutcome.passed for test in self.tests])

    def all_tests_collected_successfully(self) -> bool:
        return all([col.outcome == TestOutcome.passed for col in self.collectors])

    def non_empty_tests(self) -> bool:
        return self.summary.total > 0

    def list_errors(self) -> List[str]:
        errors = set()

        # Python errors in the function under test
        for col in self.collectors:
            if col.outcome == TestOutcome.failed and col.longrepr:
                errors.add(col.longrepr)

        # applicative test failure
        for test in self.tests:
            if test.outcome == TestOutcome.failed:
                error = test.call.longrepr
                if test.call.crash:
                    error = test.call.crash.message
                if test.user_properties:
                    case_desc = test.user_properties[0].get("docstring")
                    if case_desc:
                        error = f"""Test case {case_desc} failed with the following message:\n {test.call.crash.message}"""
                errors.add(error)
        return list(errors)

def run(folder:str, test_file:str, report_file:str)->TestReport:
    run_tests_in_safe_python_separate_process(folder, test_file, report_file)
    
    report = read_test_report(os.path.join(folder, report_file))

    #overwrite it with indented version
    with open(os.path.join(folder, report_file), "w", encoding="utf-8") as f:
        json.dump(report.model_dump(), f, indent=2)

    return report

# Run the tests in this environment.
# run the tests in safe mode, so network and os operations are blocked. only specified libraries can be used.
# run the tests in a separate process. so python modules are isolated. as the code is evolving in the filesystem, we need a way to avoid python module caching. otherwise, it will not see that the code in the file has changed.
def run_tests_in_safe_python_separate_process(folder:str, test_file:str, report_file:str):
    code = f"""
import pytest
pytest.main(["{join(folder, test_file)}", "--quiet", "--json-report", "--json-report-file={join(folder, report_file)}"])
"""   
    return run_in_process(run_safe_python, code, ["pytest"])

# def _run_in_subprocess(folder:str, test_file:str, report_file:str):
#     subprocess.run([
#             "pytest",
#             test_file,
#             # "--verbose",
#             "--quiet",
#             "--json-report", 
#             f"--json-report-file={report_file}"
#         ], 
#         env={
#             **os.environ, 
#             "PYTHONPATH": "."
#         },
#         cwd=folder)


def configure(folder: str):
    """adds the test function docstring to the output report"""

    hook = """
import pytest

def pytest_runtest_protocol(item, nextitem):
    docstring = item.function.__doc__
    if docstring:
        item.user_properties.append(("docstring", docstring))
"""
    FileTwin(file_name="conftest.py", content=hook).save(folder)


def read_test_report(file_path: str) -> TestReport:
    with open(file_path, "r") as file:
        data = json.load(file)
    return TestReport.model_validate(data, strict=False)



import re

PYTHON_FENCED_PATTERN = r"^```python\s*\n([\s\S]*)\n```"
PYTHON_UNFENCED_PATTERN = r'^python\n([\s\S]*)$'

def get_code_content(llm_code:str) -> str:
    code = llm_code.replace("\\n", "\n")
    
    match = re.match(PYTHON_FENCED_PATTERN, code)
    if match:
        return match.group(1)

    match = re.match(PYTHON_UNFENCED_PATTERN, code)
    if match:
        return match.group(1)

    return code




from typing import TypeVar
from pydantic import BaseModel, Field

from .dict import find_ref


class Reference(BaseModel):
    ref: str = Field(..., alias="$ref")


BaseModelT = TypeVar("BaseModelT", bound=BaseModel)


class DocumentWithRef(BaseModel):
    def resolve_ref(
        self, obj: Reference | BaseModelT | None, object_type: type[BaseModelT]
    ) -> BaseModelT | None:
        if isinstance(obj, Reference):
            tmp = find_ref(self.model_dump(), obj.ref)
            return object_type.model_validate(tmp)
        return obj
